/******************************************************************************
 * File: SpacesEditorModule.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesEditorModule.h"
#include "Engine/RendererSettings.h"
#include "GeneralProjectSettings.h"
#include "AndroidRuntimeSettings.h"
#include "Modules/ModuleManager.h"
#include "Widgets/Input/SButton.h"
#include "Widgets/Images/SImage.h"
#include "PropertyEditorModule.h"
#include "DetailLayoutBuilder.h"
#include "DetailCategoryBuilder.h"
#include "DetailWidgetRow.h"
#include "EditorStyleSet.h"
#include "IDetailGroup.h"
#include "ISettingsEditorModule.h"
#include "SpacesRuntimeSettings.h"
#include "Runtime/Launch/Resources/Version.h"

DEFINE_LOG_CATEGORY(LogSpacesEditor);

#define LOCTEXT_NAMESPACE "FSpacesEditorModule"

const TArray<FSpacesSettings> FSpacesSettingsDetailsCustomization::MandatorySettings {
	// START IN VR
	{
		"StartInVR",
		"Enable this setting to ensure your app starts in VR.",
		"Enable Start in VR",
		"Start in VR is enabled",
		"Please, enable the Start in VR option to ensure your app starts in VR.",
		false,
		[]()
		{
			if(UGeneralProjectSettings* const Settings = GetMutableDefault<UGeneralProjectSettings>())
			{
				Settings->bStartInVR = true;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(UGeneralProjectSettings, bStartInVR)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const UGeneralProjectSettings* const Settings = GetDefault<const UGeneralProjectSettings>())
			{
				return Settings->bStartInVR;
			}
			return false;
		}
	},
	// Vulkan is supported
	{
		"Vulkan",
		"Only the Vulkan graphics API is supported at the moment.",
		"Enable Vulkan",
		"Vulkan is enabled",
		"Please, only the Vulkan graphics API is supported at the moment.",
		false,
		[]()
		{
			if(UAndroidRuntimeSettings* const Settings = GetMutableDefault<UAndroidRuntimeSettings>())
			{
				Settings->bSupportsVulkan = true;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(UAndroidRuntimeSettings, bSupportsVulkan)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const UAndroidRuntimeSettings* const Settings = GetDefault<const UAndroidRuntimeSettings>())
			{
				return Settings->bSupportsVulkan;
			}
			return false;
		}
	},
	// SDK Version is higher than 29
	{
		"AndroidSDKVersion",
		"Minimum Android SDK version has to be equal or greater than 29.",
		"Update Android SDK version",
		"The Android SDK Version is supported",
		"Please, the minimum Android SDK version has to be equal or greater than 29.",
		false,
		[]()
		{
			if(UAndroidRuntimeSettings* const Settings = GetMutableDefault<UAndroidRuntimeSettings>())
			{
				Settings->MinSDKVersion = 29;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(UAndroidRuntimeSettings, MinSDKVersion)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const UAndroidRuntimeSettings* const Settings = GetDefault<const UAndroidRuntimeSettings>())
			{
				return Settings->MinSDKVersion >= 29;
			}
			return false;
		}
	},
	// Mobile Multi-View is enabled
	{
		"MobileMultView",
		"Enable mobile multi-view to reduce CPU usage.",
		"Enable Multi-View",
		"Multi-View is enabled",
		"Please, enable mobile multi-view to reduce CPU usage.",
		true,
		[]()
		{
			if(URendererSettings* const Settings = GetMutableDefault<URendererSettings>())
			{
				Settings->bMobileMultiView = true;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(URendererSettings, bMobileMultiView)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const URendererSettings* const Settings = GetDefault<const URendererSettings>())
			{
				return Settings->bMobileMultiView == 1;
			}
			return false;
		}
	},
	// Screen Orientation Setting
	{
		"ScreenOrientation",
		"Screen Orientation needs to be set to \"Landscape\" for correct headset rendering.",
		"Update Screen Orientation",
		"Fix Screen Orientation",
		"Change the screen orientation to \"Landscape\" for correct rendering.",
		false,
		[]()
		{
			if(UAndroidRuntimeSettings* const Settings = GetMutableDefault<UAndroidRuntimeSettings>())
			{
				Settings->Orientation = EAndroidScreenOrientation::Landscape;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(UAndroidRuntimeSettings, Orientation)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const UAndroidRuntimeSettings* const Settings = GetDefault<const UAndroidRuntimeSettings>())
			{
				return Settings->Orientation == EAndroidScreenOrientation::Landscape;
			}
			return false;
		}
	}	
};

const TArray<FSpacesSettings> FSpacesSettingsDetailsCustomization::RecommendedPerformanceSettings {
	// Allow Static Lighting
	{
		"StaticLighting",
		"Enable the setting to allow static lighting.",
		"Allow Static Lighting",
		"Static Lighting is allowed",
		"Please, allow static lighting for better performance.",
		true,
		[]()
		{
			if(URendererSettings* const Settings = GetMutableDefault<URendererSettings>())
			{
				Settings->bAllowStaticLighting = true;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(URendererSettings, bAllowStaticLighting)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const URendererSettings* const Settings = GetDefault<const URendererSettings>())
			{
				return Settings->bAllowStaticLighting == 1;
			}
			return false;
		}
	},
	// Forward shading is enabled
	{
		"ForwardShading",
		"Enable the forward shading for better performance.",
		"Enable Forward Shading",
		"Forward Shading is enabled",
		"Please, enable forward shading for better performance.",
		true,
		[]()
		{
			if(URendererSettings* const Settings = GetMutableDefault<URendererSettings>())
			{
				Settings->bForwardShading = true;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(URendererSettings, bForwardShading)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const URendererSettings* const Settings = GetDefault<const URendererSettings>())
			{
				return Settings->bForwardShading == 1;
			}
			return false;
		}
	},
	// Large aspect ratio
	{
		"AspectRatio",
		"Aspect Ratio needs to have large enough value to use the full screen for rendering.",
		"Update Aspect Ratio",
		"Aspect Ratio is valid",
		"Please, Aspect Ratio needs to have large enough value to use the full screen for rendering.",
		false,
		[]()
		{
			if(UAndroidRuntimeSettings* const Settings = GetMutableDefault<UAndroidRuntimeSettings>())
			{
				Settings->MaxAspectRatio = 10.f;
				Settings->UpdateSinglePropertyInConfigFile(Settings->GetClass()->FindPropertyByName(GET_MEMBER_NAME_CHECKED(UAndroidRuntimeSettings, MaxAspectRatio)), Settings->GetDefaultConfigFilename());
				Settings->SaveConfig();
			}
			return FReply::Handled();
		},
		[]()
		{
			if(const UAndroidRuntimeSettings* const Settings = GetDefault<const UAndroidRuntimeSettings>())
			{
				return Settings->MaxAspectRatio >= 10.f;
			}
			return false;
		}
	}
};

void FSpacesEditorModule::StartupModule()
{
	static FSpacesSettingsDetailsCustomization* Layout = new FSpacesSettingsDetailsCustomization;
	// Register settings
	FPropertyEditorModule& PropertyModule = FModuleManager::GetModuleChecked<FPropertyEditorModule>("PropertyEditor");
	PropertyModule.RegisterCustomClassLayout(USpacesRuntimeSettings::StaticClass()->GetFName(), FOnGetDetailCustomizationInstance::CreateStatic(&FSpacesSettingsDetailsCustomization::MakeInstance));


	USpacesRuntimeSettings* const SpacesSettings = GetMutableDefault<USpacesRuntimeSettings>();
	SpacesSettings->OnUpdateAndroidConfig.BindRaw(this,&FSpacesEditorModule::SetUpAndroidConfig);
	SetUpAndroidConfig(SpacesSettings->bExportHeadless);

	Layout->PreBuildLogs();	
}

void FSpacesEditorModule::ShutdownModule()
{
}

void FSpacesEditorModule::SetUpAndroidConfig(bool bExportHeadless) const
{
	UAndroidRuntimeSettings* const AndroidSettings = GetMutableDefault<UAndroidRuntimeSettings>();
	AndroidSettings->ExtraManifestNodeTags.AddUnique(TEXT("xmlns:tools=\"http://schemas.android.com/tools\""));
	const FString MergeActivity = TEXT("<activity android:name=\"com.qualcomm.snapdragon.spaces.unrealcustomlauncher.SpacesLauncher\" tools:node=\"merge\">\\n<intent-filter>\\n<category android:name=\"android.intent.category.LAUNCHER\" />\\n</intent-filter>\\n</activity>");
	const FString RemoveActivity = TEXT("<activity android:name=\"com.qualcomm.snapdragon.spaces.unrealcustomlauncher.SpacesLauncher\" tools:node=\"remove\">\\n<intent-filter>\\n<category android:name=\"android.intent.category.LAUNCHER\" />\\n</intent-filter>\\n</activity>");
	AndroidSettings->ExtraApplicationSettings.RemoveFromStart(MergeActivity);
	AndroidSettings->ExtraApplicationSettings.RemoveFromStart(RemoveActivity);
	if(bExportHeadless)
	{
		AndroidSettings->ExtraApplicationSettings.InsertAt(0,RemoveActivity);
	}
	else
	{
		AndroidSettings->ExtraApplicationSettings.InsertAt(0,MergeActivity);
	}
	AndroidSettings->SaveConfig();
	#if ENGINE_MAJOR_VERSION == 5
	AndroidSettings->TryUpdateDefaultConfigFile();
    #else
	AndroidSettings->UpdateDefaultConfigFile();
    #endif
}


void FSpacesSettingsDetailsCustomization::PreBuildLogs()
{
	for(FSpacesSettings setting : MandatorySettings)
	{
		if(!setting.IsValid())
		{
			UE_LOG(LogSpacesEditor, Error, TEXT("%s"), *setting.WarningLog);
		}
	}
	
	for(FSpacesSettings setting : RecommendedPerformanceSettings)
	{
		if(!setting.IsValid())
		{
			UE_LOG(LogSpacesEditor, Warning, TEXT("%s"), *setting.WarningLog);
		}
	}
}

TSharedRef<IDetailCustomization> FSpacesSettingsDetailsCustomization::MakeInstance()
{
	return MakeShareable(new FSpacesSettingsDetailsCustomization);
}

void FSpacesSettingsDetailsCustomization::CustomizeDetails(IDetailLayoutBuilder& DetailLayout)
{
	MyDetailLayout = nullptr;

	FSpacesSettings VerifyButton {
		"Verify",
		"Please, verify the settings always after applying any changes.",
		"Project Settings Check",
		"",
		"",
		false,
		[]()
		{
			return FReply::Handled();
		},
		[]()
		{
			return false;
		}
	};

	IDetailCategoryBuilder& RefreshCategory = DetailLayout.EditCategory("Verify Category", LOCTEXT("SpacesRefreshPerformanceCategory", ""));
	RefreshCategory.AddCustomRow(LOCTEXT("Verify Snapdragon Spaces", " "))
	.WholeRowContent()
	[
		SNew(SVerticalBox)
		+ SVerticalBox::Slot().VAlign(VAlign_Center).Padding(HPADDING_OPTION_SETTING, VPADDING_OPTION_SETTING)
		[
			CreateSettingButton(VerifyButton)
		]
	];
	
	CheckExtensionCategory(DetailLayout);
	CheckPerformanceCategory(DetailLayout);
	
	
	MyDetailLayout = &DetailLayout;
}

TSharedRef<SHorizontalBox> FSpacesSettingsDetailsCustomization::CreateSettingButton(const FSpacesSettings setting)
{
	auto HSetting = SNew(SHorizontalBox)
	+ SHorizontalBox::Slot().FillWidth(10).VAlign(VAlign_Center)
	[
		SNew(STextBlock)
		.Text(FText::Format(LOCTEXT("SpacesSettings{0}Description", "{1}"), FText::FromName(setting.Name), FText::FromString(setting.Description))).AutoWrapText(true)
	]
	+ SHorizontalBox::Slot().AutoWidth().VAlign(VAlign_Top)
	[
		SNew(SButton)
		.Text(setting.IsValid() ? FText::Format(LOCTEXT("SpacesSettingsButtonEnabled{0}", "{1}"), FText::FromName(setting.Name), FText::FromString(setting.ButtonEnabled)) : FText::Format(LOCTEXT("SpacesSettingsButtonEnable{0}", "{1}"), FText::FromName(setting.Name), FText::FromString(setting.ButtonEnable)))
		.OnClicked(this, &FSpacesSettingsDetailsCustomization::EnableSetting, setting)
		.IsEnabled(!setting.IsValid())
		
	];
	return HSetting;
}

FReply FSpacesSettingsDetailsCustomization::EnableSetting(const FSpacesSettings setting)
{
	setting.EnableSetting();
	MyDetailLayout->ForceRefreshDetails();
	AskRestart(setting.bAskRestart);
	return FReply::Handled();
}

void FSpacesSettingsDetailsCustomization::CheckExtensionCategory(IDetailLayoutBuilder& DetailLayout)
{
	
}

void FSpacesSettingsDetailsCustomization::CheckPerformanceCategory(IDetailLayoutBuilder& DetailLayout)
{
	bool bShowError = false;
	bool bShowWarning =  false;

	TSharedRef<SVerticalBox> MandatoryVBox = SNew(SVerticalBox);
	for(FSpacesSettings setting : MandatorySettings)
	{
		MandatoryVBox->AddSlot()
		.VAlign(VAlign_Center)
		.Padding(HPADDING_OPTION_SETTING, VPADDING_OPTION_SETTING)
		[
			CreateSettingButton(setting)
		];

		if(!setting.IsValid())
		{
			bShowError = true;
		}
	}

	TSharedRef<SVerticalBox> RecommendedVBox = SNew(SVerticalBox);
	for(FSpacesSettings setting : RecommendedPerformanceSettings)
	{
		RecommendedVBox->AddSlot()
		.VAlign(VAlign_Center)
		.Padding(HPADDING_OPTION_SETTING, VPADDING_OPTION_SETTING)
		[
			CreateSettingButton(setting)
			];

		if(!setting.IsValid())
		{
			bShowWarning = true;
		}
	}

	IDetailCategoryBuilder& PerformanceCategory	= DetailLayout.EditCategory("Project Settings", LOCTEXT("SpacesProjectSettingsCategory", "Project Settings"));
	PerformanceCategory
	.AddCustomRow(LOCTEXT("Project Settings Warnings Snapdragon Spaces", "Warnings"))
	.WholeRowContent()
	[
		SNew(SVerticalBox)
		+ SVerticalBox::Slot().VAlign(VAlign_Center).Padding(HPADDING_OPTION_SETTING, VPADDING_OPTION_SETTING)
		[
			SNew(SHorizontalBox)
			.Visibility((bShowError ? EVisibility::Visible : EVisibility::Collapsed))
			+ SHorizontalBox::Slot().AutoWidth().HAlign(HAlign_Center).VAlign(VAlign_Center).Padding(VPADDING_OPTION_SETTING)
			[
				SNew(SImage)
				.Image(FEditorStyle::Get().GetBrush("Icons.Error"))
			]
			+ SHorizontalBox::Slot().FillWidth(8).VAlign(VAlign_Center)
			[
				SNew(STextBlock)
				.AutoWrapText(true)
				.Text(LOCTEXT("SpacesProjectSettingsError", "ERROR: Some options should be enabled before package."))
				.ColorAndOpacity(FLinearColor(0.7f, 0.23f, 0.23f, 1.f))
				.Font(FEditorStyle::GetFontStyle(TEXT("BoldFont")))
			]
		]
		+ SVerticalBox::Slot().VAlign(VAlign_Center).Padding(HPADDING_OPTION_SETTING, VPADDING_OPTION_SETTING)
		[
			SNew(SHorizontalBox)
			.Visibility((bShowWarning ? EVisibility::Visible : EVisibility::Collapsed))
			+ SHorizontalBox::Slot().AutoWidth().HAlign(HAlign_Center).VAlign(VAlign_Center).Padding(VPADDING_OPTION_SETTING)
			[
				SNew(SImage)
				.Image(FEditorStyle::Get().GetBrush("Icons.Warning"))
			]
			+ SHorizontalBox::Slot().FillWidth(8).VAlign(VAlign_Center)
			[
				SNew(STextBlock)
				.AutoWrapText(true)
				.Text(LOCTEXT("SpacesProjectSettingsWarning", "WARNING: Some options recommended to package are pending to enable."))
				.ColorAndOpacity(FLinearColor(0.25f, 0.25f, 0.f, 1.f))
				.Font(FEditorStyle::GetFontStyle(TEXT("BoldFont")))
			]
		]
	];
	
	
	PerformanceCategory
	.AddGroup("Mandatory", LOCTEXT("Mandatory Snapdragon Spaces", "Mandatory"))
	.AddWidgetRow()
	.WholeRowContent()
	[
		MandatoryVBox
	];
	
	PerformanceCategory
	.AddGroup("Recommended for better performance", LOCTEXT("Performance Recommended Snapdragon Spaces", "Recommended"))
	.AddWidgetRow()
	.WholeRowContent()
	[
		RecommendedVBox
	];
}

void FSpacesSettingsDetailsCustomization::AskRestart(const bool required)
{
	if(!required) return;
	
	ISettingsEditorModule* SettingsEditorModule = FModuleManager::GetModulePtr<ISettingsEditorModule>("SettingsEditor");
	if (SettingsEditorModule)
	{
		SettingsEditorModule->OnApplicationRestartRequired();
	}
}

#undef LOCTEXT_NAMESPACE
	
IMPLEMENT_MODULE(FSpacesEditorModule, SpacesEditor)