/******************************************************************************
 * File: SpacesCameraAccessExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesCameraAccessExtension.h"
#include "SpacesARCameraImageTexture.h"
#if PLATFORM_ANDROID
#include "Android/AndroidPlatformMisc.h"
#endif

FSpacesCameraAccessExtension::FSpacesCameraAccessExtension()
{
	RegisterOpenXRExtensionModularFeature();
}

FSpacesCameraAccessExtension::~FSpacesCameraAccessExtension()
{
	ShutdownCamera();
	UnregisterOpenXRExtensionModularFeature();
}

UARTexture* FSpacesCameraAccessExtension::OnGetARTexture(EARTextureType TextureType) const
{
	if(!ARTextureHolder) return nullptr;
	
	if(TextureType == EARTextureType::CameraImage)
	{
		return ARTextureHolder->ARCameraRGBTexture;
	}
	return nullptr;
}

bool FSpacesCameraAccessExtension::OnGetCameraIntrinsics(FARCameraIntrinsics& OutCameraIntrinsics) const
{
	OutCameraIntrinsics = CurrentCameraIntrinsics;
	return true;
}

FString FSpacesCameraAccessExtension::GetDisplayName()
{
	return FString(TEXT("Spaces Camera Access"));
}

void FSpacesCameraAccessExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
}

void FSpacesCameraAccessExtension::PostCreateSession(XrSession InSession)
{
	SetSessionHandle(InSession);
}

bool FSpacesCameraAccessExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	OutExtensions.Add(XR_QCOMX_CAMERA_FRAME_ACCESS_EXTENSION_NAME);
	return true;
}

const void* FSpacesCameraAccessExtension::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
	return FSpacesCaptureExtension::OnCreateSession(InInstance, InSystem, InNext);
}

const void* FSpacesCameraAccessExtension::OnBeginSession(XrSession InSession, const void* InNext)
{
	SetSessionHandle(InSession);
	return FSpacesCaptureExtension::OnBeginSession(InSession, InNext);
}

void FSpacesCameraAccessExtension::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
	if(bCameraFrameConfigFound == false){ return; }
	if(UpdateFrames == false){ return; }
	UpdateCameraFrame(EARTextureType::CameraImage);
}

void FSpacesCameraAccessExtension::SetCameraAccessState(bool Active)
{
	if(!ARTextureHolder.IsValid())
	{
		UpdateFrames = false;
		return;
	}
	UpdateFrames = Active;
}

bool FSpacesCameraAccessExtension::GetCameraFrameConfigFound() const
{
	return bCameraFrameConfigFound;
}

bool FSpacesCameraAccessExtension::CanEnableExtension() const
{
#if PLATFORM_ANDROID
	const FString DeviceName = FAndroidMisc::GetDeviceModel();
	UE_LOG(LogSpacesRuntime, Warning, TEXT("Camera Access: DeviceName: %s"), *DeviceName);
	if(DeviceName.Contains("motorola edge"))
	{
		return true;
	}
#endif
	UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Feature only supported on motorola edge devices"));
	return false;
}

bool FSpacesCameraAccessExtension::OnStart()
{
	const XrSession Session = GetSessionHandle();
	CreateCameraHandles(Session);
	ARTextureHolder = MakeShared<FSpacesSharedARTextureHolder>();
	UpdateFrames = true;
	return true;
}

void FSpacesCameraAccessExtension::OnPause()
{
}

void FSpacesCameraAccessExtension::OnResume()
{
}

bool FSpacesCameraAccessExtension::OnStop()
{
	ARTextureHolder.Reset();
	ShutdownCamera();
	UpdateFrames = false;
	return true;
}

void FSpacesCameraAccessExtension::FSpacesSharedARTextureHolder::AddReferencedObjects(FReferenceCollector& Collector)
{
	Collector.AddReferencedObject(ARCameraRGBTexture);
}

bool FSpacesCameraAccessExtension::GetCamerasInfo(XrSession InSession)
{
	if(InSession == XR_NULL_HANDLE)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: XR Session is NULL."));
		return false;
	}
	
	uint32_t CamerasNumber;
	
	if(SnapdragonSpaces::xrEnumerateCamerasQCOMX(InSession, 0, &CamerasNumber, nullptr) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to enumerate cameras."));
		return false;
	}

	UE_LOG(LogSpacesRuntime, Warning, TEXT("Camera Access: Cameras Number = %d."), CamerasNumber);
	if(CamerasNumber <= 0)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: No cameras enumerated."));
		return false;
	}

	TArray<XrCameraInfoQCOMX> CameraInfoArray;
	CameraInfoArray.Init({ XR_TYPE_CAMERA_INFO_QCOMX, nullptr }, CamerasNumber);
	
	if(SnapdragonSpaces::xrEnumerateCamerasQCOMX(InSession, CamerasNumber, &CamerasNumber, CameraInfoArray.GetData()) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to enumerate cameras."));
		return false;
	}
	
	bool CameraFound = false;
	
	for(const XrCameraInfoQCOMX& CameraInfo : CameraInfoArray)
	{
		if (CameraInfo.cameraType == XR_CAMERA_TYPE_RGB_QCOMX)
		{
			CamerasInfosMap.Add(EARTextureType::CameraImage, CameraInfo);
			CameraFound = true;
		}else if(CameraInfo.cameraType == XR_CAMERA_TYPE_TRACKING_QCOMX)
		{
			CamerasInfosMap.Add(EARTextureType::EnvironmentCapture, CameraInfo);
			CameraFound = true;
		}else if(CameraInfo.cameraType == XR_CAMERA_TYPE_DEPTH_QCOMX)
		{
			CamerasInfosMap.Add(EARTextureType::CameraDepth, CameraInfo);
			CameraFound = true;
		}
	}
	
	return CameraFound;
}

XrCameraFrameConfigurationQCOMX FSpacesCameraAccessExtension::GetCamerasConfigurationInfo(XrSession InSession, const XrCameraInfoQCOMX& CurrentCameraInfo)
{
	uint32_t ConfigurationsNumber;

	XrCameraFrameConfigurationQCOMX FinalCameraConfigInfo{XR_TYPE_CAMERA_FRAME_CONFIGURATION_QCOMX, nullptr};
	
	if(SnapdragonSpaces::xrGetSupportedFrameConfigurationsQCOMX(InSession, CurrentCameraInfo.cameraSet, 0, &ConfigurationsNumber, nullptr) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to get supported frame configurations."));
		return FinalCameraConfigInfo;
	}
	
	if(ConfigurationsNumber <= 0)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: No supported frame configurations found."));
		return FinalCameraConfigInfo;
	}
	
	TArray<XrCameraFrameConfigurationQCOMX> CameraConfigArray;
	CameraConfigArray.Init({XR_TYPE_CAMERA_FRAME_CONFIGURATION_QCOMX, nullptr},ConfigurationsNumber);

	if(SnapdragonSpaces::xrGetSupportedFrameConfigurationsQCOMX(InSession, CurrentCameraInfo.cameraSet, ConfigurationsNumber, &ConfigurationsNumber, CameraConfigArray.GetData()) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to get supported frame configurations."));
		return FinalCameraConfigInfo;
	}
	
	for(const XrCameraFrameConfigurationQCOMX& CameraConfig : CameraConfigArray)
	{
		if (CameraConfig.format == XR_CAMERA_FRAME_FORMAT_YUV420_NV21_QCOMX ||
			CameraConfig.format == XR_CAMERA_FRAME_FORMAT_YUV420_NV12_QCOMX)
		{
			FinalCameraConfigInfo = CameraConfig;

			bCameraFrameConfigFound = true;

			break;
		}
	}

	return FinalCameraConfigInfo;
}

void FSpacesCameraAccessExtension::CreateCameraHandles(XrSession InSession)
{
	bCameraFrameConfigFound = GetCamerasInfo(InSession); 
	if(!bCameraFrameConfigFound)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to get all cameras info."));
		return;
	}
	
	for(const TPair<EARTextureType, XrCameraInfoQCOMX>& CameraInfo : CamerasInfosMap)
	{
		XrCameraFrameConfigurationQCOMX CameraConfigInfo = GetCamerasConfigurationInfo(InSession, CameraInfo.Value);
		if(!bCameraFrameConfigFound)
		{
			UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to get %d configuration info."), CameraInfo.Key);
			continue;
		}

		XrCameraQCOMX XrCameraHandle;
	
		XrCameraActivationFrameConfigurationInfoQCOMX CameraActivationConfigurationInfo {
			XR_TYPE_CAMERA_ACTIVATION_FRAME_CONFIGURATION_INFO_QCOMX,
			nullptr,
			CameraConfigInfo.format,
			CameraConfigInfo.resolutionName,
			CameraConfigInfo.minFPS,
			XR_FALSE
		};

		XrCameraActivationInfoQCOMX CameraActivationInfo {
			XR_TYPE_CAMERA_ACTIVATION_INFO_QCOMX,
			&CameraActivationConfigurationInfo,
			CameraInfo.Value.cameraSet
		};

		if(SnapdragonSpaces::xrCreateCameraHandleQCOMX(InSession, &CameraActivationInfo, &XrCameraHandle) == false)
		{
			UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to create camera handle."));
			continue;
		}
		
		CamerasConfigInfosMap.Add(CameraInfo.Key, CameraConfigInfo);
		XrCameraHandles.Add(CameraInfo.Key, XrCameraHandle);
	}
}

void FSpacesCameraAccessExtension::ShutdownCamera()
{
	for(const TPair<EARTextureType, XrCameraQCOMX> XrCameraHandle : XrCameraHandles)
	{
		if(SnapdragonSpaces::xrReleaseCameraHandleQCOMX(XrCameraHandle.Value) == false)
		{
			UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to destroy %d handle."), XrCameraHandle.Key);
		}
	}
}

void FSpacesCameraAccessExtension::ExtractRGBFrame(EARTextureType TextureType, TArray<XrCameraFrameBufferQCOMX>& FrameBuffers, const XrCameraSensorPropertiesQCOMX& Sensor, uint8_t*& FramePixels) const
{
	const XrCameraFrameBufferQCOMX& FrameBuffer =  FrameBuffers[0];
	
	const XrCameraFramePlaneQCOMX*  yPlane = nullptr;
	const XrCameraFramePlaneQCOMX*  uvPlane = nullptr;
	const XrCameraFramePlaneQCOMX*  Plane = FrameBuffer.planes;

	for (uint32_t planeIndex = 0; planeIndex < FrameBuffer.planeCount; ++planeIndex)
	{
		if(Plane->planeType == XR_CAMERA_FRAME_PLANE_TYPE_Y_QCOMX)
		{
			yPlane = Plane;
		}else if(Plane->planeType == XR_CAMERA_FRAME_PLANE_TYPE_UV_QCOMX)
		{
			uvPlane = Plane;
		}
		++Plane;
	}

	const bool SwapUV = CamerasConfigInfosMap[TextureType].format == XR_CAMERA_FRAME_FORMAT_YUV420_NV12_QCOMX;
	
	FramePixels = new uint8_t[Sensor.imageDimensions.width * Sensor.imageDimensions.height * 4];
	
	for(int32_t Row = 0; Row < Sensor.imageDimensions.height; ++Row)
	{
		for(int32_t Col = 0; Col < Sensor.imageDimensions.width; ++Col)
		{
			//Transforming from YUV to RGB space
			uint8_t     Y = FrameBuffer.buffer[yPlane->offset + (Sensor.imageOffset.y + Row) * yPlane->stride + Sensor.imageOffset.x + Col];
			int8_t      U = 0;
			int8_t      V = 0;

			if(uvPlane)
			{
				const auto  RowOffset = (Sensor.imageOffset.y + Row) / 2 * uvPlane->stride;
				const auto  ColOffset = (Sensor.imageOffset.x + Col) / 2 * 2;
				const auto  Offset = uvPlane->offset + RowOffset + ColOffset;

				U = static_cast< int >(FrameBuffer.buffer[Offset]) - 128;
				V = static_cast< int >(FrameBuffer.buffer[Offset + 1]) - 128;

				if (SwapUV)
				{
					Swap(U, V);
				}
			}

			auto R = Y + 1.370705f * V;
			auto G = Y - 0.698001f * V - 0.337633f * U;
			auto B = Y + 1.732446f * U;
			auto A = 255;

			R = R > 255 ? 255 : R < 0 ? 0 : R;
			G = G > 255 ? 255 : G < 0 ? 0 : G;
			B = B > 255 ? 255 : B < 0 ? 0 : B;

			//Writing in BGRA format
			const int32_t PixelIndex = ((Row * Sensor.imageDimensions.width) + Col);
			FramePixels[4 * PixelIndex] = B;
			FramePixels[4 * PixelIndex + 1] = G;
			FramePixels[4 * PixelIndex + 2] = R;
			FramePixels[4 * PixelIndex + 3] = A;
		}
	}
}

void FSpacesCameraAccessExtension::UpdateCameraFrame(EARTextureType TextureType)
{
	if(!ARTextureHolder.IsValid()) return;
	
	if(!XrCameraHandles.Contains(TextureType))
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to get camera handle while getting camera frame."));
		return;
	}
	
	const XrCameraQCOMX& CameraHandle = XrCameraHandles[TextureType];

	TArray<XrCameraFrameBufferQCOMX> FrameBuffers;
	if(!CamerasConfigInfosMap.Contains(TextureType))
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to get Camera config info."));
		return;
	}
	FrameBuffers.Init({ XR_TYPE_CAMERA_FRAME_BUFFER_QCOMX, nullptr },CamerasConfigInfosMap[TextureType].frameBufferCount);

	TArray<XrCameraSensorPropertiesQCOMX> Sensors;
	if(!CamerasInfosMap.Contains(TextureType))
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to get Camera info."));
		return;
	}
	Sensors.Init({ XR_TYPE_CAMERA_SENSOR_PROPERTIES_QCOMX, nullptr }, CamerasInfosMap[TextureType].sensorCount);

	XrCameraSensorInfosQCOMX    CameraSensorInfos {XR_TYPE_CAMERA_SENSOR_INFOS_QCOMX,nullptr,static_cast< uint32_t >(Sensors.Num()),Sensors.GetData()};

	XrCameraFrameBuffersQCOMX   CameraFrameBuffers {XR_TYPE_CAMERA_FRAME_BUFFERS_QCOMX,&CameraSensorInfos,static_cast< uint32_t >(FrameBuffers.Num()),FrameBuffers.GetData()};

	XrCameraFrameDataQCOMX CameraFrameData { XR_TYPE_CAMERA_FRAME_DATA_QCOMX, &CameraFrameBuffers };

	if(SnapdragonSpaces::xrAccessFrameQCOMX(CameraHandle, &CameraFrameData) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to access frame."));
		return;
	}
	
	int SensorID = 0;
	for(const auto& Sensor: Sensors)
	{
		if(SensorID == SelectedCameraSensorId)
		{
			UpdateCameraIntrinsics(Sensor);
			if(TextureType == EARTextureType::CameraImage)
			{
				uint8_t* PixelsData = nullptr;
				//Supporting only color camera (for now)
				ExtractRGBFrame(TextureType, FrameBuffers, Sensor, PixelsData);
				if(ARTextureHolder->ARCameraRGBTexture == nullptr)
				{
					ARTextureHolder->ARCameraRGBTexture = UARTexture::CreateARTexture<USpacesARCameraImageTexture>(EARTextureType::CameraImage);
				}
				ARTextureHolder->ARCameraRGBTexture->UpdateFrameRGBTexture(PixelsData, FVector2D(Sensor.imageDimensions.width, Sensor.imageDimensions.height));
				delete[] PixelsData;
			}
			break;
		}
		++SensorID;
	}

	if(SnapdragonSpaces::xrReleaseFrameQCOMX(CameraFrameData.handle) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Camera Access: Failed to release frame."));
	}
}

void FSpacesCameraAccessExtension::UpdateCameraIntrinsics(const XrCameraSensorPropertiesQCOMX& Sensor)
{
	CurrentCameraIntrinsics.ImageResolution = FIntPoint(Sensor.imageDimensions.width, Sensor.imageDimensions.height);
	CurrentCameraIntrinsics.PrincipalPoint = FVector2D(Sensor.intrinsics.principalPoint.x, Sensor.intrinsics.principalPoint.y);
	CurrentCameraIntrinsics.FocalLength = FVector2D(Sensor.intrinsics.focalLength.x, Sensor.intrinsics.focalLength.y);
}

bool FSpacesCameraAccessExtension::GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes)
{
	OutTypes.Add(EARCaptureType::Camera);
	return true;
}

bool FSpacesCameraAccessExtension::OnEnabledCaptureType(EARCaptureType InType)
{
	return true;
}

bool FSpacesCameraAccessExtension::OnDisabledCaptureType(EARCaptureType InType)
{
	return true;
}

#if PLATFORM_ANDROID
void FSpacesCameraAccessExtension::GetRequiredPermissions(TArray<FString>& Permissions) const
{
	Permissions.Add("android.permission.CAMERA");
}
#endif