/******************************************************************************
 * File: SpacesHandTrackingExtension.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#pragma once

#include "CoreMinimal.h"
#include "SpacesExtension.h"
#include "HeadMountedDisplayTypes.h"
#include "IHandTracker.h"
#include "IXRTrackingSystem.h"
#include "XRMotionControllerBase.h"

class IOpenXRARTrackedMeshHolder;

class FSpacesHandTrackingExtension : public FSpacesCaptureExtension, public FXRMotionControllerBase, public IHandTracker
{
public:
	struct FHandState : public FNoncopyable
	{
		FHandState();

		XrHandTrackerEXT HandTracker{};
		XrHandJointLocationEXT HandJointLocations[XR_HAND_JOINT_COUNT_EXT];
		XrHandJointLocationsEXT HandLocations{ XR_TYPE_HAND_JOINT_LOCATIONS_EXT };
		
		FTransform HandKeypointTransforms[EHandKeypointCount];
		float Radii[EHandKeypointCount];
		bool JointPosesAvailable;

		EARTrackingState HandTrackingState = EARTrackingState::Unknown;
		
		//Hand gestures
		XrHandGestureQCOM HandGestureData;

		//Hand Meshing
		XrSpace HandMeshSpace{};
		TArray<uint32_t> HandMeshIndices;
		TArray<XrHandMeshVertexMSFT> HandMeshVertices;

		FGuid HandMeshGuid;
		
		int HandMeshIndicesNum;
		int HandMeshVerticesNum;
		int HandMeshIndicesMaxNum;
		int HandMeshVerticesMaxNum;

		FTransform HmLocalToTrackingTransform;
		
		bool GetTransform(EHandKeypoint KeyPoint, FTransform& OutTransform) const;
		const FTransform& GetTransform(EHandKeypoint KeyPoint) const;
	};
	
	FSpacesHandTrackingExtension();
	virtual ~FSpacesHandTrackingExtension() override;

	/** IOpenXRExtensionPlugin implementation */
	virtual FString GetDisplayName() override;
	virtual void PostCreateInstance(XrInstance InInstance) override;
	virtual void PostCreateSession(XrSession InSession) override;
	virtual bool GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions) override;
	virtual const void* OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext) override;
	virtual const void* OnBeginSession(XrSession InSession, const void* InNext) override;
	virtual void UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace) override;

	/** IMotionController interface */
	virtual bool GetControllerOrientationAndPosition(const int32 ControllerIndex, const FName MotionSource, FRotator& OutOrientation, FVector& OutPosition, float WorldToMetersScale) const override;
	virtual bool GetControllerOrientationAndPosition(const int32 ControllerIndex, const EControllerHand DeviceHand, FRotator& OutOrientation, FVector& OutPosition, float WorldToMetersScale) const override;
	virtual ETrackingStatus GetControllerTrackingStatus(const int32 ControllerIndex, const EControllerHand DeviceHand) const override;
	virtual FName GetMotionControllerDeviceTypeName() const override;
	virtual void EnumerateSources(TArray<FMotionControllerSource>& SourcesOut) const override;

	/** IHandTracker */
	virtual FName GetHandTrackerDeviceTypeName() const override;
	virtual bool IsHandTrackingStateValid() const override;
	virtual bool GetKeypointState(EControllerHand Hand, EHandKeypoint Keypoint, FTransform& OutTransform, float& OutRadius) const override;
	virtual bool GetAllKeypointStates(EControllerHand Hand, TArray<FVector>& OutPositions, TArray<FQuat>& OutRotations, TArray<float>& OutRadii) const override;

	virtual bool HasHandMeshData() const override;
	virtual bool GetHandMeshData(EControllerHand Hand, TArray<FVector>& OutVertices, TArray<FVector>& OutNormals, TArray<int32>& OutIndices, FTransform& OutHandMeshTransform) const override;
	
	/** SpacesHandTrackingModule */
	const FHandState& GetLeftHandState() const;
	const FHandState& GetRightHandState() const;
	bool GetHandTrackingSupported() const;
	const XrHandGestureQCOM GetHandGestureData(int HandIndex) const;
	bool HandStateJointPosesAvailable(int HandIndex) const;
	void SetHandMeshingStatus(bool Active);
	bool GetHandMeshingStatus() const;
	void SetHandTrackingState(bool active);

	/** SpacesCaptureExtension implementation */
	virtual bool GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes) override;
	
protected:
	
	virtual bool OnEnabledCaptureType(EARCaptureType InType) override;
	virtual bool OnDisabledCaptureType(EARCaptureType InType) override;
	virtual bool CanEnableExtension() const override;
	virtual bool OnStart() override;
	virtual void OnPause() override;
	virtual void OnResume() override;
	virtual bool OnStop() override;

#if PLATFORM_ANDROID
	virtual bool IsRuntimeCameraPermissionsRequired() const override {return true;};
#endif
	
private:
	
	IXRTrackingSystem* XRTrackingSystem = nullptr;

	bool bHandTrackingSupported = false;
	
	bool bHandMeshingActive = false;

	bool bHandTrackingActive = false;

	int32 HandDeviceIndex;

	FHandState HandStates[2];

	IOpenXRARTrackedMeshHolder* HandsTrackedMeshHolder = nullptr;
	
	void UpdateHandTracking(int HandIndex, XrTime DisplayTime, XrSpace TrackingSpace);
	void TryGetHandGestureData(int HandIndex, XrTime DisplayTime);
	void UpdateHandMeshing(int HandIndex, XrTime DisplayTime, XrSpace TrackingSpace);
	
	bool LocateAndUpdateHandMesh(FHandState& HandState, XrTime DisplayTime, XrSpace TrackingSpace);
	void FillHandMeshData(FHandState& HandState);
	
	void InitHandState(XrSession InSession, int HandIndex);
	void InitHandGestures();

	void RemoveHandMesh(const FHandState& HandState);
	void RemoveHandMeshes(); 
};


