﻿/******************************************************************************
 * File: SpacesHitTestingExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesHitTestingExtension.h"
#include "ARBlueprintLibrary.h"
#include "OpenXRCore.h"
#include "OpenXRHMD/Private/OpenXRHMD.h"
#include "SpacesRuntimeBlueprintLibrary.h"
#include "SpacesRuntimeModule.h"


#define LOCTEXT_NAMESPACE "FSpacesHitTestingExtension"


FSpacesHitTestingExtension::FSpacesHitTestingExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Startup SpacesHitTestingExtension"));
	RegisterOpenXRExtensionModularFeature();
}

FSpacesHitTestingExtension::~FSpacesHitTestingExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Shutdown SpacesHitTestingExtension"));
	UnregisterOpenXRExtensionModularFeature();
}

FString FSpacesHitTestingExtension::GetDisplayName()
{
	return FString(TEXT("Hit Testing"));
}

void FSpacesHitTestingExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
}

void FSpacesHitTestingExtension::PostCreateSession(XrSession InSession)
{
    SetSessionHandle(InSession);
	
    /* Inspect supported ray casting target types */
    if(!SnapdragonSpaces::xrEnumerateRayCastTargetTypesQCOM.IsValid())
    {
        UE_LOG(LogSpacesRuntime, Error, TEXT("xrEnumerateRayCastTargetTypesQCOM method not found"));
    }

    uint32_t TypeCount = 0;
    if(SnapdragonSpaces::xrEnumerateRayCastTargetTypesQCOM(GetSessionHandle(), 0, &TypeCount, nullptr) == false)
    {
        UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to enumerate ray cast target types"));
    }
	
    SupportedTargetTypes.SetNum(TypeCount, true);
    if(SnapdragonSpaces::xrEnumerateRayCastTargetTypesQCOM(GetSessionHandle(), TypeCount, &TypeCount, SupportedTargetTypes.GetData()) == false)
    {
        UE_LOG(LogSpacesRuntime, Error, TEXT("xrEnumerateRayCastTargetTypesQCOM fails"));
    }

    for(const auto Type : SupportedTargetTypes)
    {
        UE_LOG(LogSpacesRuntime, Log, TEXT("Ray Cast supports the target type %s"), SnapdragonSpaces::XrEnumStr(Type));
    }

	//Creating raycast handle at the begginning
	XrRayCastCreateInfoQCOM CreateInfo = {
    	XrStructureType::XR_TYPE_RAY_CAST_CREATE_INFO_QCOM,
		nullptr,
		XrRayTypeQCOM::XR_RAY_TYPE_STOPPING_QCOM,
		static_cast<uint32_t>(SupportedTargetTypes.Num()),
		SupportedTargetTypes.GetData()
	};
	Ray = SnapdragonSpaces::FRayCastQCOMHandle(GetSessionHandle(), &CreateInfo);
}

const void* FSpacesHitTestingExtension::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
    XRTrackingSystem = GEngine->XRSystem.Get();
	return InNext;
}

bool FSpacesHitTestingExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	OutExtensions.Add(XR_QCOM_RAY_CASTING_EXTENSION_NAME);
	return true;
}

bool FSpacesHitTestingExtension::CanEnableExtension() const
{
	return SupportedTargetTypes.Num() > 0;
}

bool FSpacesHitTestingExtension::OnStart()
{
	return true;
}

void FSpacesHitTestingExtension::OnPause()
{
}

void FSpacesHitTestingExtension::OnResume()
{
}

bool FSpacesHitTestingExtension::OnStop()
{
	return true;
}

TArray<FARTraceResult> FSpacesHitTestingExtension::OnLineTraceTrackedObjects(const TSharedPtr<FARSupportInterface, ESPMode::ThreadSafe> ARCompositionComponent, const FVector Start, const FVector End, const EARLineTraceChannels TraceChannels)
{
    //TODO(AS): Enable/disable target types based on EARLineTraceChannels AND enabled features
	
	if(!Ray.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("xrCreateRayCastQCOM fails"));
		return TArray<FARTraceResult>();
	}
	
	/* Cast into current view direction */

	if(!SnapdragonSpaces::xrCastRayQCOM.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("xrCastRayQCOM method not found"));
		return TArray<FARTraceResult>();
	}

	const XrSpace TrackingSpace = GetXRSystem()->GetTrackingSpace();
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();

	const FTransform WorldToTracking = XRTrackingSystem->GetTrackingToWorldTransform().Inverse();
	FVector RayOriginInCm = WorldToTracking.TransformPosition(Start);
	FVector RayEndInCm = WorldToTracking.TransformPosition(End);
	FVector Direction = (RayEndInCm - RayOriginInCm);
	
	XrVector3f RayOrigin = ToXrVector(RayOriginInCm, WorldToMetersScale);
	XrVector3f RayDirection = ToXrVector(Direction, WorldToMetersScale);
	float RayDistance = (Direction.Size() / WorldToMetersScale);
	constexpr float MaxRayDistance = 42.0f;
	RayDistance = (RayDistance < MaxRayDistance) ? RayDistance : MaxRayDistance;

	if(SnapdragonSpaces::xrCastRayQCOM(Ray.Value(), TrackingSpace, RayOrigin, RayDirection, RayDistance) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed Cast Ray"));
		return TArray<FARTraceResult>();
	}
	/* Check the ray cast result */

	if(!SnapdragonSpaces::xrGetRayCastStateQCOM.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("xrGetRayCastStateQCOM method not found"));
		return TArray<FARTraceResult>();
	}
	
	XrRayCastStateQCOM RayCastState;
	if(SnapdragonSpaces::xrGetRayCastStateQCOM(Ray.Value(), &RayCastState) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to get RayCast State!"));
		return TArray<FARTraceResult>();
	}

	if(RayCastState != XR_RAY_CAST_STATE_FINISHED_QCOM)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("RayCastState = %s"), SnapdragonSpaces::XrEnumStr(RayCastState));
		return TArray<FARTraceResult>();
	}

	if(!SnapdragonSpaces::xrGetRayCollisionsQCOM.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("xrGetRayCollisionsQCOM method not found"));
		return TArray<FARTraceResult>();
	}

	XrRayCollisionsGetInfoQCOM GetInfo {
		XR_TYPE_RAY_COLLISIONS_GET_INFO_QCOM,
		nullptr,
		TrackingSpace
	};

	uint32_t CollisionCountOutput = 0;
	XrRayCollisionsQCOM Collisions {
		XR_TYPE_RAY_COLLISIONS_QCOM,
		nullptr,
		0,
		&CollisionCountOutput,
		nullptr
	};

	if(SnapdragonSpaces::xrGetRayCollisionsQCOM(Ray.Value(), &GetInfo, &Collisions) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to get RayCast collisions capacity output!"));
		return TArray<FARTraceResult>();
	}

	if(*Collisions.collisionCountOutput == 0)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Collisions.collisionCountOutput = %u"), *Collisions.collisionCountOutput);
		return TArray<FARTraceResult>();
	}
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("Collisions.collisionCapacityInput = %u"), *Collisions.collisionCountOutput);
		
	TArray<XrRayCollisionQCOM> CollisionsBuffer;
	CollisionsBuffer.Init({XR_TYPE_RAY_COLLISION_QCOM}, *Collisions.collisionCountOutput);
	Collisions.collisionCapacityInput = CollisionsBuffer.Num();
	Collisions.collisions = CollisionsBuffer.GetData();
	
	if(SnapdragonSpaces::xrGetRayCollisionsQCOM(Ray.Value(), &GetInfo, &Collisions) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to get RayCast collisions! (%s)"), SnapdragonSpaces::XrEnumStr(SnapdragonSpaces::xrGetRayCollisionsQCOM.GetLastResult()));
		CollisionsBuffer.Empty();
		return TArray<FARTraceResult>();
	}
	
	if(*Collisions.collisionCountOutput == 0) return TArray<FARTraceResult>();
	
	TArray<FARTraceResult> Results;
	Results.Reserve(*Collisions.collisionCountOutput);

	const FTransform TrackingToWorldTransform = XRTrackingSystem->GetTrackingToWorldTransform();
	const auto TrackedGeometries = UARBlueprintLibrary::GetAllGeometries();
	
	for (uint32_t i = 0; i < *Collisions.collisionCountOutput; ++i)
	{
		XrRayCollisionQCOM Collision = Collisions.collisions[i];
		const FVector HitPoint = ToFVector(Collision.position, WorldToMetersScale);
		const FVector HitNormal = ToFVector(Collision.surfaceNormal, WorldToMetersScale);
		const FTransform HitLocalTransform(HitNormal.ToOrientationQuat(), HitPoint);

		const FTransform HitWorldTransform = HitLocalTransform * TrackingToWorldTransform;
		const float Distance = (HitWorldTransform.GetLocation() - RayOriginInCm).Size();

		UARTrackedGeometry* Geometry = nullptr;
		if(const auto GeometryPtr = TrackedGeometries.FindByPredicate(
			[id=Collision.targetId](const UARTrackedGeometry* Current){ return Current->UniqueId.D == id;}
			); GeometryPtr != nullptr)
		{
			Geometry = *GeometryPtr;
		}
	    
		UE_LOG(LogSpacesRuntime, Log, TEXT("Collision.targetId = %llu (distance: %f)"), Collision.targetId, Distance);
		Results.Add(FARTraceResult(ARCompositionComponent, Distance, TraceChannels, HitLocalTransform, Geometry));
	}
	
	CollisionsBuffer.Empty();
	UE_LOG(LogSpacesRuntime, Log, TEXT("Raytrace is hitting with %u planes"), Results.Num());
	return Results;
}

bool FSpacesHitTestingExtension::GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes)
{
	OutTypes.Append({EARCaptureType::SpatialMapping, EARCaptureType::SceneUnderstanding});
	return true;
}

bool FSpacesHitTestingExtension::OnEnabledCaptureType(EARCaptureType InType)
{
	return true;
}

bool FSpacesHitTestingExtension::OnDisabledCaptureType(EARCaptureType InType)
{
	return true;
}

#if PLATFORM_ANDROID
void FSpacesHitTestingExtension::GetRequiredPermissions(TArray<FString>& Permissions) const
{
	Permissions.Add("android.permission.CAMERA");
}
#endif

#undef LOCTEXT_NAMESPACE
