/******************************************************************************
 * File: SpacesImageTrackingExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesImageTrackingExtension.h"
#include "SpacesCommon.h"
#include "OpenXRCore.h"
#include "IXRTrackingSystem.h"
#include "Engine/Texture2D.h"
#include "Async/Async.h"
#include "SubSystems/SpacesEngineSubsystem.h"

#define LOCTEXT_NAMESPACE "FSpacesImageTrackingExtension"


const XrImageTrackerDataSetBaseHeaderQCOM* const* FTrackedImagesHelper::GetDataSets() const
{
	return reinterpret_cast<const XrImageTrackerDataSetBaseHeaderQCOM* const*>(TrackedDataSetImages.GetData());
}

int32_t FTrackedImagesHelper::GetDataSetsCount() const
{
	return TrackedDataSetImages.Num();
}

UARCandidateImage* FTrackedImagesHelper::FindCandidateImageById(const FString& Id) const
{
	UARCandidateImage* const * Result = TrackedCandidateImages.Find(Id);
	if(Result == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Image not found for given Id (%s)"), *Id);
		return nullptr;
	}
	return *Result;
}

int32_t FTrackedImagesHelper::GetCandidateImageCount() const
{
	return TrackedCandidateImages.Num();
}

bool FTrackedImagesHelper::TryGetRGBImagesFromTextures(TArray<UTexture2D*> Textures, TArray<int32> NameIndex, TArray<float> IPHeights)
{
	if(!(Textures.Num() == NameIndex.Num() && Textures.Num() == IPHeights.Num()))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("TryGetRGBImagesFromTextures, arrays size do not match!"));
		return false;
	}

	for(int k = 0; k < Textures.Num(); ++k)
	{
		FRGBImageData ResultData;
		
		ResultData.ImageNameIndex = NameIndex[k];
		ResultData.IPHeight = IPHeights[k];
			
		if(Textures[k] == nullptr)
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Texture cannot be NULL, index: %d"), k);
			continue;
		}
			
		if(Textures[k]->GetPixelFormat() != PF_B8G8R8A8)
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Invalid texture pixel format, must be BGRA, index: %d"), k);
			continue;
		}
	
		auto Mip0 = Textures[k]->GetPlatformMips()[0];
	
		if(Mip0.SizeX * Mip0.SizeY * Mip0.SizeZ * 4 != Mip0.BulkData.GetBulkDataSize())
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Invalid Mip0, incorrect bulk data size or resolution. Index: %d"), k);
			continue;
		}
		
		const uint32_t BufferSize = Mip0.SizeX * Mip0.SizeY * Mip0.SizeZ * 3;
		uint8_t* Buffer = new uint8_t[BufferSize];
		uint8_t* OrigBuffer = new uint8_t[Mip0.BulkData.GetBulkDataSize()];
		Mip0.BulkData.GetCopy(reinterpret_cast<void**>(&OrigBuffer), false);

		for(int j = 0; j < Mip0.SizeY; ++j)
		{
			for(int i = 0; i < Mip0.SizeX; ++i)
			{
				const int pixel = i + j * Mip0.SizeX;
				const int pixelFlip = i + (Mip0.SizeY - j - 1) * Mip0.SizeX;
				 			
				const uint8_t* OriginalPixel = &OrigBuffer[pixel*4];
				uint8_t* ConvertedPixel = &Buffer[pixelFlip*3];
				ConvertedPixel[0] = OriginalPixel[2];
				ConvertedPixel[1] = OriginalPixel[1];
				ConvertedPixel[2] = OriginalPixel[0];
			}
		}
		delete [] OrigBuffer;
				
		ResultData.RGBImage = Buffer;
		ResultData.RGBImageSize = BufferSize;
		ResultData.PixelSize.x = Mip0.SizeX;
		ResultData.PixelSize.y = Mip0.SizeY;
		
		CreateTrackedDataSetImage(ResultData);
	}
	return true;
}

FTrackedImagesHelper::FTrackedImagesHelper(const TArray<UARCandidateImage*>& CandidateImages)
{
	
	TArray<UTexture2D*> Textures;
	TArray<int32> NameIndices;
	TArray<float> ImageHeights;
	
	for(const auto CandidateImage : CandidateImages)
	{
		if(CandidateImage == nullptr)
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Null ARCandidate Image"));
			continue;
		}
		
		const float IPWidth = CandidateImage->GetPhysicalWidth();
		const float IPHeight = CandidateImage->GetPhysicalHeight();

		if(IPWidth <= 0.0f || IPHeight <= 0.0f)
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Physical width and height should be greater than 0.0"));
			continue;
		}

		if(CandidateImage->GetFriendlyName() == "")
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("ARCandidateImage: Warning! FriendlyName field is empty."));
		}
		
		UTexture2D* Texture = CandidateImage->GetCandidateTexture();
		TArray<ANSICHAR> Name(TCHAR_TO_ANSI(*CandidateImage->GetFriendlyName()), CandidateImage->GetFriendlyName().Len()+1);
		const auto NameIndex = NameCandidateImages.Add(MoveTemp(Name));
		
		Textures.Add(Texture);
		NameIndices.Add(NameIndex);
		ImageHeights.Add(IPHeight);
		
		TrackedCandidateImages.Add(CandidateImage->GetFriendlyName(), CandidateImage);
	}
	
	TPromise<bool> Promise;
	ImagesLoadedFuture = Promise.GetFuture();
	
	AsyncTask(
		ENamedThreads::AnyThread,
		[Promise = MoveTemp(Promise), TexturesArray = Textures, NamesIndicesArray = NameIndices, ImageHeightsArray = ImageHeights, this]() mutable 
		{
			Promise.SetValue(TryGetRGBImagesFromTextures(TexturesArray, NamesIndicesArray, ImageHeightsArray));
		}
	);
	
}

void FTrackedImagesHelper::CreateTrackedDataSetImage(FRGBImageData ImageData)
{
	XrImageTrackerDataSetImageQCOM* Data = new XrImageTrackerDataSetImageQCOM
		{
			XrStructureType::XR_TYPE_IMAGE_TRACKER_DATA_SET_IMAGE_QCOM,
			nullptr,
			NameCandidateImages[ImageData.ImageNameIndex].GetData(),
			ImageData.IPHeight / 100, // the physical height in meters of the reference image.
			ImageData.PixelSize,
			ImageData.RGBImage,
			ImageData.RGBImageSize
		};
	TrackedDataSetImages.Add(Data);
}

bool FTrackedImagesHelper::CheckTrackedDataSetImagesCreation() const 
{
	if(ImagesLoadedFuture.IsReady())
	{
		const FString ImagesLoadingResult = ImagesLoadedFuture.Get() ? "Success" : "Fail";
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Images loading result: %s"), *ImagesLoadingResult);
		return true;
	}
	
	return false;
}

void FTrackedImagesHelper::ResetImagesLoadedFuture()
{
	ImagesLoadedFuture.Reset();
}

FTrackedImagesHelper::~FTrackedImagesHelper()
{
	
	ImagesLoadedFuture.Reset();
	
	for(const auto TrackedDataSetImage : TrackedDataSetImages)
	{
		delete [] TrackedDataSetImage->buffer;
		delete TrackedDataSetImage;
	}
}

UARTrackedGeometry* FSpacesOpenXRImageData::ConstructNewTrackedGeometry(TSharedPtr<FARSupportInterface, ESPMode::ThreadSafe> ARSupportInterface)
{
	check(IsInGameThread());

	UARTrackedQRCode* NewImage = NewObject<UARTrackedQRCode>();
	NewImage->UniqueId = Id;
	NewImage->SetObjectClassification(EARObjectClassification::Image);
	return NewImage;
}

void FSpacesOpenXRImageData::UpdateTrackedGeometry(UARTrackedGeometry* TrackedGeometry, TSharedPtr<FARSupportInterface, ESPMode::ThreadSafe> ARSupportInterface)
{
	
	check(IsInGameThread());

	UARTrackedImage* UpdatedImage = Cast<UARTrackedImage>(TrackedGeometry);
	check(UpdatedImage != nullptr);

	UpdatedImage->UpdateTrackedGeometry(ARSupportInterface.ToSharedRef(),
		GFrameCounter,
		Timestamp,
		LocalToTrackingTransform,
		ARSupportInterface->GetAlignmentTransform(),
		Size,
		DetectedImage);
	UpdatedImage->SetTrackingState(TrackingState);
	
}

bool FSpacesImageTrackingExtension::TryCreateImageTracker(bool SupportsImageTracking)
{
	if (!SnapdragonSpaces::xrCreateImageTrackerQCOM.IsValid())
	{
		return false;
	}
		

	// Set optimization hints
	if (SnapdragonSpaces::xrSetTrackingOptimizationSettingsHintQCOM.IsValid())
	{
		if (bLowPowerMode) 
		{
			SnapdragonSpaces::xrSetTrackingOptimizationSettingsHintQCOM(
				GetSessionHandle(),
				XrTrackingOptimizationSettingsDomainQCOM::XR_TRACKING_OPTIMIZATION_SETTINGS_DOMAIN_IMAGE_TRACKING_QCOM,
				XrTrackingOptimizationSettingsHintQCOM::XR_TRACKING_OPTIMIZATION_SETTINGS_HINT_LOW_POWER_PRIORIZATION_QCOM
			);
		}

		if (bExtendedRangeMode) 
		{
			SnapdragonSpaces::xrSetTrackingOptimizationSettingsHintQCOM (
				GetSessionHandle(),
				XrTrackingOptimizationSettingsDomainQCOM::XR_TRACKING_OPTIMIZATION_SETTINGS_DOMAIN_IMAGE_TRACKING_QCOM,
				XrTrackingOptimizationSettingsHintQCOM::XR_TRACKING_OPTIMIZATION_SETTINGS_HINT_LONG_RANGE_PRIORIZATION_QCOM
			);
		}
	}
	
	// Create Image Tracker
	XrImageTrackerCreateInfoQCOM CreateInfo{
		XR_TYPE_IMAGE_TRACKER_CREATE_INFO_QCOM,
		nullptr,
		(uint32_t)TrackedDataSets->GetDataSetsCount(),
		TrackedDataSets->GetDataSets(),
		(uint32_t)CurrentSessionConfig->GetMaxNumSimultaneousImagesTracked()
	};

	MaxNumSimultaneousImagesTracked = CreateInfo.imageCount;
	UE_LOG(LogSpacesRuntime, Log, TEXT("MaxNumSimultaneousImagesTracked == %u"), MaxNumSimultaneousImagesTracked);

	
	ImageTracker = SnapdragonSpaces::FImageTrackerQCOMHandle(GetSessionHandle(), &CreateInfo);
	
	if (!ImageTracker.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to create Image Tracker"));
		return false;
	}
	
	return true;
}

bool FSpacesImageTrackingExtension::TryDestroyImageTracker()
{
	ImageTracker.Clear();
	return !ImageTracker.IsValid();
}

bool FSpacesImageTrackingExtension::TryLocateImageTargets(XrSpace Space, XrTime Time, TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>> &LocatedImages)
{
	LocatedImages.Reset();
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("TryLocateImageTargets started"));
	if (!SnapdragonSpaces::xrLocateImageTargetsQCOM.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("XrLocateImageTargetsQCOM method not found!"));
		return false;
	}

	XrImageTargetsLocateInfoQCOM LocateInfo
	{
		 XR_TYPE_IMAGE_TARGETS_LOCATE_INFO_QCOM,
		 nullptr,
		 Space,
		 Time
	};

	TArray<XrImageTargetLocationQCOM> ImageTargetLocations;
	ImageTargetLocations.SetNum(MaxNumSimultaneousImagesTracked);

	XrImageTargetLocationsQCOM ImageLocations
	{
		XR_TYPE_IMAGE_TARGET_LOCATIONS_QCOM,
		nullptr,
		0,
		MaxNumSimultaneousImagesTracked,
		ImageTargetLocations.GetData()
	};

	if (SnapdragonSpaces::xrLocateImageTargetsQCOM(ImageTracker.Value(), &LocateInfo, &ImageLocations) == false) return false;

	if (ImageLocations.imageCount == 0)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Not detected images to locate"));
		return true;
	}

	if (ImageLocations.isActive)
	{
		for (uint32_t i = 0; i < ImageLocations.imageCount; ++i)
		{
			const XrPosef& imageInWorld = ImageLocations.imageTargetLocations[i].pose;
			const XrImageTargetLocationQCOM& ImageLocation = ImageLocations.imageTargetLocations[i];
			
			// convert handle to string/number pair
			XrImageTargetQCOM imageTarget = ImageLocation.imageTarget;
			FString identifier = "";
			uint32_t id = 0;
			// FFileHelper::SaveArrayToFile(TArray64<uint8_t>(((XrImageTrackerDataSetImageQCOM*)TrackedDataSets->GetDataSets()[0])->buffer, ((XrImageTrackerDataSetImageQCOM*)TrackedDataSets->GetDataSets()[0])->bufferSize), TEXT("/sdcard/dump.1.raw"));
			if (TryGetImageTargetAndId(imageTarget, &id, &identifier))
			{
				if(UARCandidateImage* CandidateImage = TrackedDataSets->FindCandidateImageById(identifier))
				{
					TSharedPtr<FSpacesOpenXRImageData> XRImage = MakeShared<FSpacesOpenXRImageData>();
					XRImage->DetectedImage = CandidateImage;
					XRImage->Size = FVector2D(CandidateImage->GetPhysicalWidth(), CandidateImage->GetPhysicalHeight());
					XRImage->Timestamp = FPlatformTime::Seconds();
					XRImage->Id = FGuid(0, 0, 0, id);
					XRImage->TargetLocationQCOM = ImageLocation;

					LocatedImages.Add(XRImage->Id, XRImage);
				}
			}
		}
	}
	
	return true;
}

bool FSpacesImageTrackingExtension::TryGetImageTargetAndId(XrImageTargetQCOM imageTarget, uint32_t* id, FString* identifier) const
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("TryGetImageTargetAndId started"));
	if (!SnapdragonSpaces::xrImageTargetToNameAndIdQCOM.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("XrImageTargetToNameAndIdQCOM method not found!"));
		return false;
	}

	
	uint32_t nameBufferCapacity;
	if (SnapdragonSpaces::xrImageTargetToNameAndIdQCOM(imageTarget, 0, &nameBufferCapacity, nullptr, nullptr) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to locate image targets"));
		return false;
	}

	if(nameBufferCapacity == 0)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("xrImageTargetToNameAndIdQCOM: name buffer capacity is 0"));
		return false;
	}

	TArray<ANSICHAR> nameBuffer;
	nameBuffer.SetNum(nameBufferCapacity);
	uint32_t nameBufferBytesWritten;
	if (SnapdragonSpaces::xrImageTargetToNameAndIdQCOM(imageTarget, nameBufferCapacity, &nameBufferBytesWritten, nameBuffer.GetData(), id) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to locate image targets"));
		return false;
	}
	
	if (nameBuffer[nameBufferCapacity-1] == '\0')
	{
		nameBufferBytesWritten -= 1;
	}
	
	*identifier = FString(nameBufferBytesWritten, nameBuffer.GetData());

	UE_LOG(LogSpacesRuntime, Log, TEXT("Tracked target: %s"), **identifier);
	
	return true;
}

void FSpacesImageTrackingExtension::CreateImageTrackerDataSetImages()
{
	TrackedDataSets = MakeUnique<FTrackedImagesHelper>(CurrentSessionConfig->GetCandidateImageList());
}

FSpacesImageTrackingExtension::FSpacesImageTrackingExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Startup SpacesImageTrackingExtension"));
	RegisterOpenXRExtensionModularFeature();
}

FSpacesImageTrackingExtension::~FSpacesImageTrackingExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Shutdown SpacesImageTrackingExtension started"));
	ImageTrackerCreationFuture.Reset();
	TryDestroyImageTracker();
	UnregisterOpenXRExtensionModularFeature();
}

const void* FSpacesImageTrackingExtension::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnCreateSession started"));
	// Inspect image tracking system properties
	XrSystemImageTrackingPropertiesQCOM SystemImageTrackingProperties{
		XR_TYPE_SYSTEM_IMAGE_TRACKING_PROPERTIES_QCOM
	};

	XrSystemProperties SystemProperties{
		XR_TYPE_SYSTEM_PROPERTIES,
		&SystemImageTrackingProperties
	};

	if (XR_SUCCEEDED(OpenXRDynamicAPI::xrGetSystemProperties(InInstance, InSystem, &SystemProperties)))
	{
		bDeviceSupportsImageTracking = SystemImageTrackingProperties.supportsImageTracking == XR_TRUE;
		UE_LOG(LogSpacesRuntime, Log, TEXT("bDeviceSupportsImageTracking == %s"),bDeviceSupportsImageTracking?TEXT("True"):TEXT("False"));
		//TODO(AS):Force it for now
		bDeviceSupportsImageTracking = true;
	}
	
	XRTrackingSystem = GEngine->XRSystem.Get();
	
	ImageHolder = IModularFeatures::Get().GetModularFeatureImplementations<IOpenXRARTrackedGeometryHolder>(IOpenXRARTrackedGeometryHolder::GetModularFeatureName())[0];
	
	return InNext;
}

void FSpacesImageTrackingExtension::PostCreateSession(XrSession InSession)
{
	SetSessionHandle(InSession);
}

void FSpacesImageTrackingExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
}

FString FSpacesImageTrackingExtension::GetDisplayName()
{
	return FString(TEXT("Spaces Image Tracking"));
}

bool FSpacesImageTrackingExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	OutExtensions.Add(XR_QCOM_IMAGE_TRACKING_EXTENSION_NAME);
	return true;
}

void FSpacesImageTrackingExtension::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
	if (!IsEnabled())
	{
		return;
	}

	if(!bTrackerCreated)
	{
		if(TrackedDataSets->CheckTrackedDataSetImagesCreation() && !bCreateTrackerCalled)
		{
			TPromise<bool> CreateTrackerPromise;
			ImageTrackerCreationFuture = CreateTrackerPromise.GetFuture();
			AsyncTask(
					ENamedThreads::AnyThread,[Promise = MoveTemp(CreateTrackerPromise), this]() mutable
					{
						Promise.SetValue(TryCreateImageTracker(true));
					}
			);
			bCreateTrackerCalled = true;
		}
		
		if(ImageTrackerCreationFuture.IsReady())
		{
			bTrackerCreated = ImageTrackerCreationFuture.Get();
			if(USpacesEngineSubsystem* SpacesSubsystem = GEngine->GetEngineSubsystem<USpacesEngineSubsystem>())
			{
				SpacesSubsystem->OnSpacesImageTrackingIsReady.Broadcast(bTrackerCreated);
			}
			const FString CreationResult = bTrackerCreated ? "True" : "False" ;
			UE_LOG(LogSpacesRuntime, Warning, TEXT("CreateImageTrackerResult: %s "), *CreationResult);
		}
		return;
	}
	
	TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>> LocatedImages;
	if(!TryLocateImageTargets(TrackingSpace, DisplayTime, LocatedImages)) return;
	
	TArray<FGuid> MissingImagesId;
	CalculateMissingImages(LocatedImages, MissingImagesId);
	
	RemoveImages(MissingImagesId);
	UpdateImages(LocatedImages);

	UE_LOG(LogSpacesRuntime, Log, TEXT("UpdateDeviceLocations success. Updated Images: %d"), PreviousTrackedImages.Num());
}

bool FSpacesImageTrackingExtension::GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes)
{
	OutTypes.Add(EARCaptureType::SceneUnderstanding);
	return true;
}

bool FSpacesImageTrackingExtension::OnEnabledCaptureType(EARCaptureType InType)
{
	return true;
}

bool FSpacesImageTrackingExtension::OnDisabledCaptureType(EARCaptureType InType)
{
	return true;
}

void FSpacesImageTrackingExtension::CalculateMissingImages(const TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>>& LocatedImages, TArray<FGuid>& MissingImagesId)
{
	for(auto previous : PreviousTrackedImages)
	{
		if(!LocatedImages.Contains(previous.Key))
		{
			MissingImagesId.Add(previous.Key);
		}
	}
}

void FSpacesImageTrackingExtension::RemoveImages(const TArray<FGuid>& MissingImagesId)
{
	if(ImageHolder == nullptr) return;
	
	for (auto id : MissingImagesId)
	{
		if (PreviousTrackedImages.Contains(id))
		{
			ImageHolder->ARTrackedGeometryRemoved(*PreviousTrackedImages.Find(id));
			PreviousTrackedImages.Remove(id);
		}
	}
}

void FSpacesImageTrackingExtension::UpdateImages(const TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>>& Images)
{
	if(ImageHolder == nullptr) return;
	
	const float WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();

	for (auto image : Images)
	{
		if(!PreviousTrackedImages.Contains(image.Key))
		{
			ImageHolder->ARTrackedGeometryAdded(image.Value);
		}
		
		TSharedPtr<FSpacesOpenXRImageData> ImageData = image.Value;
		ImageData->Timestamp = FPlatformTime::Seconds();

		XrImageTargetLocationQCOM TargetLocationQCOM = ImageData->TargetLocationQCOM;
		if (SnapdragonSpaces::IsValidSpaceLocation(TargetLocationQCOM.locationFlags))
		{
			ImageData->LocalToTrackingTransform = ToFTransform(TargetLocationQCOM.pose, WorldToMetersScale);
			ImageData->TrackingState = EARTrackingState::Tracking;
		}
		else
		{
			ImageData->TrackingState = EARTrackingState::NotTracking;
		}

		ImageHolder->ARTrackedGeometryUpdated(ImageData);
		PreviousTrackedImages.Add(image.Key, ImageData);
	}
	
}

void FSpacesImageTrackingExtension::ClearImages() const
{
	for(auto image : PreviousTrackedImages)
	{
		ImageHolder->ARTrackedGeometryRemoved(image.Value);
	}
}

bool FSpacesImageTrackingExtension::CanEnableExtension() const
{
	return bDeviceSupportsImageTracking
	&& ImageHolder != nullptr
    && CurrentSessionConfig->GetCandidateImageList().Num() > 0
	;
}

bool FSpacesImageTrackingExtension::OnStart()
{
	ClearImages();
	CreateImageTrackerDataSetImages();
	return true;
}

void FSpacesImageTrackingExtension::OnPause()
{
}

void FSpacesImageTrackingExtension::OnResume()
{
}

bool FSpacesImageTrackingExtension::OnStop()
{
	ImageTracker.Clear();
	bTrackerCreated = false;
	bCreateTrackerCalled = false;
	ImageTrackerCreationFuture.Reset();
	TrackedDataSets->ResetImagesLoadedFuture();
	return TryDestroyImageTracker();
}

#if PLATFORM_ANDROID
void FSpacesImageTrackingExtension::GetRequiredPermissions(TArray<FString>& Permissions) const
{
	Permissions.Add("android.permission.CAMERA");
}
#endif


#undef LOCTEXT_NAMESPACE