/******************************************************************************
 * File: SpacesImageTrackingExtension.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#pragma once

#include "CoreMinimal.h"
#include "External/openxr/openxr.h"
#include "SpacesExtension.h"
#include "IOpenXRARTrackedGeometryHolder.h"

#include <string>

struct FSpacesOpenXRImageData : public FOpenXRQRCodeData
{

	UARCandidateImage* DetectedImage = nullptr;
	XrImageTargetLocationQCOM TargetLocationQCOM;

	FSpacesOpenXRImageData(): TargetLocationQCOM()
	{
	}

	virtual UARTrackedGeometry* ConstructNewTrackedGeometry(TSharedPtr<FARSupportInterface, ESPMode::ThreadSafe> ARSupportInterface) override;
	virtual void UpdateTrackedGeometry(UARTrackedGeometry* TrackedGeometry, TSharedPtr<FARSupportInterface, ESPMode::ThreadSafe> ARSupportInterface) override;
};

struct FRGBImageData
{
	uint8_t* RGBImage;
	uint32_t RGBImageSize;
	XrVector2f PixelSize;
	int32 ImageNameIndex;
	float IPHeight = 0.0;
};

class FTrackedImagesHelper
{
public:
	/***
	 *	Returns a pointer to the first element of the XrImageTrackerDataSetBaseHeaderQCOM* array
	 */
	const XrImageTrackerDataSetBaseHeaderQCOM* const* GetDataSets() const;
	int32_t GetDataSetsCount() const;
	
	/***
	 *	Returns a pointer to the CandidateImage or nullptr if none found
	 */
	UARCandidateImage* FindCandidateImageById(const FString& Id) const;
	int32_t GetCandidateImageCount() const;
	
	void CreateTrackedDataSetImage(FRGBImageData ImageData);
	bool CheckTrackedDataSetImagesCreation() const;

	void ResetImagesLoadedFuture();
	
    explicit FTrackedImagesHelper(const TArray<UARCandidateImage*>& CandidateImages);
	~FTrackedImagesHelper();
	
private:
	TArray<TArray<ANSICHAR>> NameCandidateImages;
	TArray<XrImageTrackerDataSetImageQCOM*> TrackedDataSetImages;
	TMap<FString, UARCandidateImage*> TrackedCandidateImages;

	TFuture<bool> ImagesLoadedFuture;

	bool bCreateDataSetCalled = false;
	
	bool TryGetRGBImagesFromTextures(TArray<UTexture2D*> Textures, TArray<int32> NameIndex, TArray<float> IPHeights);
};

class IXRTrackingSystem;
class FSpacesImageTrackingExtension : public FSpacesCaptureExtension
{
	
public:

	FSpacesImageTrackingExtension();
	virtual ~FSpacesImageTrackingExtension() override;

	/** IOpenXRExtensionPlugin implementation */
	virtual const void* OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext) override;
	virtual void PostCreateSession(XrSession InSession) override;
	virtual void PostCreateInstance(XrInstance InInstance) override;
	virtual FString GetDisplayName() override;
	virtual bool GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions) override;
	virtual void UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace) override;

protected:	
#if PLATFORM_ANDROID
	virtual void GetRequiredPermissions(TArray<FString>& Permissions) const override;
#endif
	
	/** SpacesCaptureExtension implementation */
	public:
	virtual bool GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes) override;
protected:
	virtual bool OnEnabledCaptureType(EARCaptureType InType) override;
	virtual bool OnDisabledCaptureType(EARCaptureType InType) override;
	virtual bool CanEnableExtension() const override;
	virtual bool OnStart() override;
	virtual void OnPause() override;
	virtual void OnResume() override;
	virtual bool OnStop() override;

private:

	TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>> PreviousTrackedImages;
	TArray<FString> CandidateImageIDs;
	TUniquePtr<FTrackedImagesHelper> TrackedDataSets;
	
	SnapdragonSpaces::FImageTrackerQCOMHandle ImageTracker;
	
	IOpenXRARTrackedGeometryHolder* ImageHolder = nullptr;
	IXRTrackingSystem* XRTrackingSystem = nullptr;

	uint32_t MaxNumSimultaneousImagesTracked = 0;
	
	bool bDeviceSupportsImageTracking = false;
	
	bool bExtendedRangeMode = false;
	bool bLowPowerMode = false;

	bool bTrackerCreated = false;

	bool bCreateTrackerCalled = false;

	TFuture<bool> ImageTrackerCreationFuture;

	bool TryCreateImageTracker(bool SupportsImageTracking);
	bool TryDestroyImageTracker();

	bool TryLocateImageTargets(XrSpace Space, XrTime Time, TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>> &LocatedImages);
	bool TryGetImageTargetAndId(XrImageTargetQCOM imageTarget, uint32_t* id, FString* identifier) const;
	FORCEINLINE void CreateImageTrackerDataSetImages();

	
	void CalculateMissingImages(const TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>>& LocatedImages, TArray<FGuid>& MissingImagesId);
	void RemoveImages(const TArray<FGuid>& MissingImagesId);
	void UpdateImages(const TMap<FGuid, TSharedPtr<FSpacesOpenXRImageData>>& NewImages);
	void ClearImages() const;

};
