﻿/******************************************************************************
 * File: JavaClassObjectExtended.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "JavaClassObjectExtended.h"

#include "SpacesCommon.h"

#if PLATFORM_ANDROID
#include "Android/AndroidApplication.h"

#if USE_ANDROID_JNI
#include "Android/AndroidJNI.h"

FScopedJavaObject<jobject> AndroidJavaEnv::GetApplicationContextThis()
{
    UE_LOG(LogSpacesRuntime, Log, TEXT("Getting Application Context..."));
    JNIEnv*	JEnv = GetJavaEnv();
    jmethodID geApplicationContextMethod = JEnv->GetMethodID(FJavaWrapper::GameActivityClassID, "getApplicationContext", "()Landroid/content/Context;");
    return NewScopedJavaObject(JEnv, JEnv->CallObjectMethod(FJavaWrapper::GameActivityThis, geApplicationContextMethod));
}

jfieldID FJavaClassObjectExtended::FindField(const char* FieldName, const char* FieldType)
{
    JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
    jfieldID Field = JEnv->GetFieldID(Class, FieldName, FieldType);
    // Is field valid?
    checkf(Field, TEXT("Unable to find Java Field %s with Type %s"), UTF8_TO_TCHAR(FieldName), UTF8_TO_TCHAR(FieldType));
    return Field;
}

FJavaClassMethod FJavaClassObjectExtended::GetStaticClassMethod(const char* MethodName, const char* FuncSig)
{
    JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
    FJavaClassMethod Method;
    Method.Method = JEnv->GetStaticMethodID(Class, MethodName, FuncSig);
    Method.Name = MethodName;
    Method.Signature = FuncSig;
    // Is method valid?
    checkf(Method.Method, TEXT("Unable to find Java Static Method %s with Signature %s"), UTF8_TO_TCHAR(MethodName), UTF8_TO_TCHAR(FuncSig));
    return Method;
}


/**** Field Getters ****/
template<>
bool FJavaClassObjectExtended::GetField<bool>(jfieldID FieldID)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    bool RetVal =  JEnv->GetBooleanField(Object, FieldID);
    VerifyException();
    return RetVal;
}

template<>
int FJavaClassObjectExtended::GetField<int>(jfieldID FieldID)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    int RetVal =  JEnv->GetIntField(Object, FieldID);
    VerifyException();
    return RetVal;
}

template<>
jobject FJavaClassObjectExtended::GetField<jobject>(jfieldID FieldID)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    jobject val =  JEnv->GetObjectField(Object, FieldID);
    VerifyException();
    jobject RetVal = JEnv->NewGlobalRef(val);
    JEnv->DeleteLocalRef(val);
    return RetVal;
}

template<>
jobjectArray FJavaClassObjectExtended::GetField<jobjectArray>(jfieldID FieldID)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    jobject val =  JEnv->GetObjectField(Object, FieldID);
    VerifyException();
    jobjectArray RetVal = (jobjectArray)JEnv->NewGlobalRef(val);
    JEnv->DeleteLocalRef(val);
    return RetVal;
}

template<>
int64 FJavaClassObjectExtended::GetField<int64>(jfieldID FieldID)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    int64 RetVal =  JEnv->GetLongField(Object, FieldID);
    VerifyException();
    return RetVal;
}

template<>
FString FJavaClassObjectExtended::GetField<FString>(jfieldID FieldID)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    jstring val = static_cast<jstring>(
        JEnv->GetObjectField(Object, FieldID));
    VerifyException();
    FString RetVal = FJavaHelper::FStringFromLocalRef(JEnv, val);
    return RetVal;
}


/**** Field Setters ****/
template<>
void FJavaClassObjectExtended::SetField<bool>(jfieldID FieldID, const bool& Value)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    JEnv->SetBooleanField(Object, FieldID, Value);
    VerifyException();
}

template<>
void FJavaClassObjectExtended::SetField<int>(jfieldID FieldID, const int& Value)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    JEnv->SetIntField(Object, FieldID, Value);
    VerifyException();
}

template<>
void FJavaClassObjectExtended::SetField<jobject>(jfieldID FieldID, const jobject& Value)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    JEnv->SetObjectField(Object, FieldID, Value);
    VerifyException();
}

template<>
void FJavaClassObjectExtended::SetField<jobjectArray>(jfieldID FieldID, const jobjectArray& Value)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    JEnv->SetObjectField(Object, FieldID, Value);
    VerifyException();
}

template<>
void FJavaClassObjectExtended::SetField<int64>(jfieldID FieldID, const int64& Value)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    JEnv->SetLongField(Object, FieldID, Value);
    VerifyException();
}

template<>
void FJavaClassObjectExtended::SetField<FString>(jfieldID FieldID, const FString& Value)
{
    JNIEnv* JEnv = FAndroidApplication::GetJavaEnv();
    auto string = FJavaClassObject::GetJString(Value);
    JEnv->SetObjectField(Object, FieldID, *string);
    VerifyException();
}


/**** StaticMethod Calls ****/
template<>
void FJavaClassObjectExtended::CallStaticMethod<void>(FJavaClassMethod Method, ...)
{
	JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
	va_list Params;
	va_start(Params, Method);
	JEnv->CallStaticVoidMethodV(Class, Method.Method, Params);
	va_end(Params);
	VerifyException();
}

template<>
bool FJavaClassObjectExtended::CallStaticMethod<bool>(FJavaClassMethod Method, ...)
{
	JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
	va_list Params;
	va_start(Params, Method);
	bool RetVal = JEnv->CallStaticBooleanMethodV(Class, Method.Method, Params);
	va_end(Params);
	VerifyException();
	return RetVal;
}

template<>
int FJavaClassObjectExtended::CallStaticMethod<int>(FJavaClassMethod Method, ...)
{
	JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
	va_list Params;
	va_start(Params, Method);
	int RetVal = JEnv->CallStaticIntMethodV(Class, Method.Method, Params);
	va_end(Params);
	VerifyException();
	return RetVal;
}

template<>
jobject FJavaClassObjectExtended::CallStaticMethod<jobject>(FJavaClassMethod Method, ...)
{
	JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
	va_list Params;
	va_start(Params, Method);
	jobject val = JEnv->CallStaticObjectMethodV(Class, Method.Method, Params);
	va_end(Params);
	VerifyException();
	jobject RetVal = JEnv->NewGlobalRef(val);
	JEnv->DeleteLocalRef(val);
	return RetVal;
}

template<>
jobjectArray FJavaClassObjectExtended::CallStaticMethod<jobjectArray>(FJavaClassMethod Method, ...)
{
	JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
	va_list Params;
	va_start(Params, Method);
	jobject val = JEnv->CallStaticObjectMethodV(Class, Method.Method, Params);
	va_end(Params);
	VerifyException();
	jobjectArray RetVal = (jobjectArray)JEnv->NewGlobalRef(val);
	JEnv->DeleteLocalRef(val);
	return RetVal;
}

template<>
int64 FJavaClassObjectExtended::CallStaticMethod<int64>(FJavaClassMethod Method, ...)
{
	JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
	va_list Params;
	va_start(Params, Method);
	int64 RetVal = JEnv->CallStaticLongMethodV(Class, Method.Method, Params);
	va_end(Params);
	VerifyException();
	return RetVal;
}

template<>
FString FJavaClassObjectExtended::CallStaticMethod<FString>(FJavaClassMethod Method, ...)
{
	JNIEnv*	JEnv = AndroidJavaEnv::GetJavaEnv();
	va_list Params;
	va_start(Params, Method);
	jstring RetVal = static_cast<jstring>(
		JEnv->CallStaticObjectMethodV(Class, Method.Method, Params));
	va_end(Params);
	VerifyException();
	auto Result = FJavaHelper::FStringFromLocalRef(JEnv, RetVal);
	return Result;
}


#endif //USE_ANDROID_JNI
#endif //PLATFORM_ANDROID