﻿/******************************************************************************
 * File: SpacesMeshingExtension.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#pragma once

#include "CoreMinimal.h"
#include "IOpenXRARModule.h"
#include "IOpenXRARTrackedGeometryHolder.h"
#include "External/openxr/openxr.h"
#include "SpacesExtension.h"


namespace SnapdragonSpaces
{
	FORCEINLINE EARObjectClassification ToARObjectClassification(XrSceneObjectTypeMSFT Type);
	FORCEINLINE EARPlaneOrientation ToARPlaneOrientation(XrScenePlaneAlignmentTypeMSFT Type);
}

class IXRTrackingSystem;
class FSpacesMeshingExtension : public FSpacesCaptureExtension
{

public:

	FSpacesMeshingExtension();
	virtual ~FSpacesMeshingExtension() override;

	/** IOpenXRExtensionPlugin implementation */
	
	virtual const void* OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext) override;
	virtual void PostCreateSession(XrSession InSession) override;
	virtual void PostCreateInstance(XrInstance InInstance) override;
	virtual FString GetDisplayName() override;
	virtual bool GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions) override;
	virtual void UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace) override;
	
	/** SpacesCaptureExtension implementation */
	public:
	virtual bool GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes) override;
protected:
	virtual bool OnEnabledCaptureType(EARCaptureType InType) override;
	virtual bool OnDisabledCaptureType(EARCaptureType InType) override;
	virtual bool CanEnableExtension() const override;
	virtual bool OnStart() override;
	virtual void OnPause() override;
	virtual void OnResume() override;
	virtual bool OnStop() override;

public:

	TArray<FTrackedMeshHolder> GetMeshes();

	static TArray<FVector> ComputeVerticesNormals(TArray<FVector> Vertices, TArray<MRMESH_INDEX_TYPE> Indices, TArray<FVector>& Normals);
	void AsyncComputeVerticesNormals();

	TArray<FVector> NormalsTest;

	void SetComputeNormalsState(bool Active);
	bool GetComputeNormalsState() const;
	
protected:
#if PLATFORM_ANDROID
	virtual bool IsRuntimeCameraPermissionsRequired() const override {return true;};
#endif
	
private:

	enum class ECurrentStep
	{
		ComputeScene,
		ComputingScene,
		LoadScene,
		UpdateScene
	};
	
	TMap<FGuid, FTrackedPlaneHolder> PlanesMap;
	TMap<FGuid, FTrackedMeshHolder> MeshesMap;
	TMap<FGuid, FTrackedMeshHolder> CollidersMap;

	TArray<FGuid> PreviousPlanes;
	TArray<FGuid> PreviousMeshes;
	TArray<FGuid> PreviousColliders;
	
	TArray<XrSceneComputeFeatureMSFT> AvailableSceneComputeFeatures;
	TArray<XrSceneComputeFeatureMSFT> EnabledSceneComputeFeatures;

	TArray<XrSceneOrientedBoxBoundMSFT> SceneBoxBounds;
	TArray<XrSceneSphereBoundMSFT> SceneSphereBounds;

	XrNewSceneComputeInfoMSFT SceneComputeInfo;

	SnapdragonSpaces::FSceneObserverMSFTHandle SceneObserver;
	
	FDateTime NextComputeSceneTime;
	
	IXRTrackingSystem* XRTrackingSystem;
	IOpenXRARTrackedMeshHolder* TrackedMeshHolder;
	
	ECurrentStep CurrentUpdateStep = ECurrentStep::ComputeScene;

	TArray<FTrackedMeshHolder> MeshesArray;

	bool bDeviceSupportsMeshing = false;

	TFuture<void> NormalsComputationFuture;

	int PreviousIndicesNum = 0;

	bool bLockVertices = false;

	bool bComputeNormals = true;

	bool TryCreateSceneObserver();
	void CalculateEnabledSceneComputeFeatures();
	
	bool TryReadSceneMeshBuffers(XrSceneMSFT SceneHandle, uint64_t MeshBufferId, XrBool32 Supports16BitIndices,
	                          TArray<FVector>& Vertices, TArray<MRMESH_INDEX_TYPE>& Indices) const;

	template<XrSceneComponentTypeMSFT Value, typename Type>
	bool TryGetComponents(XrSceneMSFT Scene, TArray<XrSceneComponentMSFT>& SceneComponents,
	                   TArray<Type>& TypeComponents) const;

	bool TryGenerateObjectMap(XrSceneMSFT Scene, TMap<FGuid, XrSceneObjectMSFT>& ObjectMap) const;
	bool TryGeneratePlanesMap(XrSceneMSFT Scene, const TMap<FGuid, XrSceneObjectMSFT>& ObjectMap);
	bool TryGenerateMeshesMap(XrSceneMSFT Scene, const TMap<FGuid, XrSceneObjectMSFT>& ObjectMap);
	bool TryGenerateCollidersMap(XrSceneMSFT Scene, const TMap<FGuid, XrSceneObjectMSFT>& ObjectMap);

	bool TryLocateUuids(XrTime DisplayTime, XrSpace TrackingSpace, XrSceneMSFT Scene);

	void RemovePlanes(const TArray<FGuid>& MissingPlanesId) const;
	void RemoveMeshes(const TArray<FGuid>& MissingMeshesId) const;

	void ClearAll();
	
	void ComputeScene(XrTime DisplayTime, XrSpace TrackingSpace);
	void ComputingScene();
	void LoadScene(XrTime DisplayTime, XrSpace TrackingSpace);
	void UpdateScene();
	
};
