﻿/******************************************************************************
 * File: SpacesCommon.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#pragma once

#include "CoreMinimal.h"
#include "External/openxr/openxr_qcom.h"

#include "HAL/Platform.h"

#if PLATFORM_ANDROID
#include "Android/AndroidApplication.h"
#if USE_ANDROID_JNI
#include "Android/AndroidJNI.h"
#endif
#endif

#include "External/openxr/openxr_platform.h"
#include "External/openxr/openxr_reflection.h"

#include "Core/Public/Misc/ConfigCacheIni.h"
#include "Core/Public/Logging/LogMacros.h"

#include "ARTypes.h"

#include "Runtime/Launch/Resources/Version.h"

DECLARE_LOG_CATEGORY_EXTERN(LogSpacesRuntime, Log, All)

/**
 * @brief Reads settings defined in the Engine.ini file
 * @param Module Module name of the class where the property is defined
 * @param Class Class name (without the prefix) where the property is defined
 * @param Name Name of the property
 * @param Type Type of the property, valid types are: Str, Int, Float, Double, Bool, Array, SingleLineArray, Color, Vector2D, Vector, Vector4 and Rotator
 * @param DefaultValue Fallback value in case the property is not found in the settings
 * @return Found value or default
 */
#define READ_CONFIG_VALUE(Module, Class, Name, Type, DefaultValue)				\
	[](const decltype(DefaultValue)& Default){									\
		checkf(GConfig->IsReadyForUse(), TEXT("GConfig is not initialized."));	\
		decltype(DefaultValue) OutputValue;										\
		if(!GConfig->Get##Type(													\
			TEXT("/Script/"#Module"."#Class""),									\
			TEXT(#Name),														\
			OutputValue,														\
			GEngineIni															\
		)){																		\
			return Default;														\
		}																		\
		return OutputValue;														\
	}(DefaultValue);


//From openxr.h
#ifndef XR_ENUM_CASE_STR
#define XR_ENUM_CASE_STR(name, val) case name: return TEXT(#name);
#endif
#define XR_ENUM_STR(enumType)								\
	constexpr const TCHAR* XrEnumStr(enumType e) {			\
		switch (e) {										\
			XR_LIST_ENUM_##enumType(XR_ENUM_CASE_STR)		\
			default: return TEXT("Unknown");				\
		}													\
	}														


namespace SnapdragonSpaces
{
	XR_ENUM_STR(XrResult);
	XR_ENUM_STR(XrStructureType);
	XR_ENUM_STR(XrSceneObjectTypeMSFT);
	XR_ENUM_STR(XrSceneComputeStateMSFT);
	XR_ENUM_STR(XrSceneComputeFeatureMSFT);
	XR_ENUM_STR(XrRayCastStateQCOM);
	XR_ENUM_STR(XrRayCastTargetTypeQCOM);
	XR_ENUM_STR(XrPlaneDetectionStateQCOM);
	XR_ENUM_STR(XrSceneComponentTypeMSFT);
	XR_ENUM_STR(XrEnvironmentBlendMode);

	template<typename XREnum>
	FString ToString(XREnum Value) {					
		return XrEnumStr(Value);								
	}				

}

#undef XR_ENUM_CASE_STR
#undef XR_ENUM_STR


enum class ESpacesExtension : uint8;
namespace SnapdragonSpaces
{
	constexpr const TCHAR* BoolToStr(const bool Value)
	{
		return Value?TEXT("True"):TEXT("False");
	}
	
	constexpr bool IsValidSpaceLocation(const XrSpaceLocationFlags Flags)
	{
		constexpr XrSpaceLocationFlags ValidFlags = (
			XR_SPACE_LOCATION_POSITION_VALID_BIT |
			XR_SPACE_LOCATION_ORIENTATION_VALID_BIT);
		return (ValidFlags & Flags) == ValidFlags;
	}
	/***
	 * @brief Cast one type to another
	 * @details Reinterpret cast one type to another ensuring they have the same size.
	 * It casts by reference, the returned value is a reference to the original.
	 * @tparam To Type to be converted to
	 * @tparam From Original type of the variable
	 * @param Value Value to convert
	 * @return Type converted
	 */
	template<typename To, typename From>
	FORCEINLINE constexpr To& As(From& Value){
		static_assert(sizeof(To) == sizeof(From), "Types have different sizes!");
		return reinterpret_cast<To&>(Value);
	};

	/***
	 * @brief Cast one pointer type to another
	 * @details Reinterpret cast one pointer type to another ensuring they have the same size
	 * @tparam To Type to be converted to
	 * @tparam From Original type of the variable
	 * @param Value Pointer to convert
	 * @return Type converted pointer
	 */
	template<typename To, typename From>
	FORCEINLINE constexpr To* As(From* Value){
		static_assert(sizeof(To) == sizeof(From), "Types have different sizes!");
		return reinterpret_cast<To*>(Value);
	};
	
	constexpr bool IsTrackedSpaceLocation(const XrSpaceLocationFlags Flags)
	{
		constexpr XrSpaceLocationFlags TrackedFlags = (
			XR_SPACE_LOCATION_POSITION_TRACKED_BIT |
			XR_SPACE_LOCATION_ORIENTATION_TRACKED_BIT);
		return (TrackedFlags & Flags) == TrackedFlags;
	}

	FORCEINLINE FGuid SpacesGuid(const ESpacesExtension Extension, uint32 Id)
	{
		return FGuid((uint32)Extension, 0, 0, Id);
	}

	FORCEINLINE void GuidToExtensionAndId(const FGuid& Guid, ESpacesExtension& Extension, uint32& Id)
	{
		Extension = (ESpacesExtension)Guid.A;
		Id = Guid.D;
	}

	constexpr EARTrackingState ToARTrackingState(XrSpaceLocationFlags Flags)
	{
		if(!IsValidSpaceLocation(Flags)){ return EARTrackingState::StoppedTracking;}
		if(!IsTrackedSpaceLocation(Flags)){ return EARTrackingState::NotTracking; }
		return EARTrackingState::Tracking;
	}
	
}


