/******************************************************************************
 * File: SpacesCoreExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesCoreExtension.h"
#include "SpacesCommon.h"
#include "SpacesComponentVersion.h"

#if PLATFORM_ANDROID
#include <dlfcn.h>
#include <android_native_app_glue.h>
#include "JNI/RuntimeCheckerWrapper.h"
#include "Misc/MessageDialog.h"
#endif //PLATFORM_ANDROID

#include "OpenXRCore.h"
#include "CustomCaptures/SpacesSceneUnderstanding.h"
#include "CustomCaptures/SpacesSpatialMapping.h"
#include "CustomCaptures/SpacesCameraCapture.h"
#include "CustomCaptures/SpacesHandMeshCapture.h"

#include "SpacesOpenXRIntercept.h"

#include "time.h"

#define LOCTEXT_NAMESPACE "FSpacesCoreExtension"

XrTime FSpacesCoreExtension::XrTimeFromDateTime(const FDateTime& DateTime) const
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("GetXrTimeFromDateTime (%s)"), *(DateTime.ToString()));

	if(!SnapdragonSpaces::xrConvertTimespecTimeToTimeKHR.IsValid()) return 0;
		    
	XrTime OutTime = 0;
	const timespec InTime{
		(time_t)DateTime.ToUnixTimestamp(),
		(int32_t)((DateTime.GetTicks() % 10000000) * 100)
	}; 

	UE_LOG(LogSpacesRuntime, Verbose, TEXT("In Time %lld (s) %ld (ns)"), InTime.tv_sec, InTime.tv_nsec);
    
	const bool Result = SnapdragonSpaces::xrConvertTimespecTimeToTimeKHR(GetInstanceHandle(), &InTime, &OutTime);

	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Out Time %lld"), OutTime);
    
	if(Result == false) return 0;

	return OutTime;	
}

FDateTime FSpacesCoreExtension::DateTimeFromXrTime(const XrTime& XRTime) const
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("GetDateTimeFromXrTime (%lld)"), XRTime);
    
	if(!SnapdragonSpaces::xrConvertTimeToTimespecTimeKHR.IsValid()) return FDateTime::MinValue();
	
	timespec OutTime;
	    
	const bool Result = SnapdragonSpaces::xrConvertTimeToTimespecTimeKHR(GetInstanceHandle(), XRTime, &OutTime);

	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Out Time %lld (s) %ld (ns)"), OutTime.tv_sec, OutTime.tv_nsec);
    
	if(Result == false) return FDateTime::MinValue();
	
	const FDateTime ConvertedTime = FDateTime::FromUnixTimestamp(OutTime.tv_sec) + FTimespan(static_cast<int64>(OutTime.tv_nsec / 100));
	
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("GetXrTimeFromDateTime output (%s)"), *(ConvertedTime.ToString()));

	return ConvertedTime;
	
}

bool FSpacesCoreExtension::GetComponentVersions(TArray<FSpacesComponentVersion>& ComponentVersions) const
{
	if (!SnapdragonSpaces::xrGetComponentVersionsQCOM.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Invalid xrGetComponentVersionsQCOM"));
		return false;
	}

	uint32_t ComponentVersionCount;
	if (SnapdragonSpaces::xrGetComponentVersionsQCOM(GetInstanceHandle(), 0, &ComponentVersionCount, nullptr) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Unsupported xrGetComponentVersionsQCOM"));
		return false;
	}

	if (ComponentVersionCount == 0)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Validation Failure"))
		return true;
	}

	TArray<XrComponentVersionQCOM> ComponentVersionsQCOM;
	ComponentVersionsQCOM.Init({XR_TYPE_COMPONENT_VERSION_QCOM}, ComponentVersionCount);

	if (SnapdragonSpaces::xrGetComponentVersionsQCOM(GetInstanceHandle(), ComponentVersionCount, &ComponentVersionCount, ComponentVersionsQCOM.GetData()) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Unsupported xrGetComponentVersionsQCOM"));
		return false;
	}

	if (ComponentVersionsQCOM.Num() == 0)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("ComponentVersions is empty"));
		return false;
	}

	FString logCV = "Enumerating component information from Spaces Services:\n";
	for(const XrComponentVersionQCOM& ComponentVersion : ComponentVersionsQCOM)
	{
		FSpacesComponentVersion SpacesComponentVersion
		{
			FString(UTF8_TO_TCHAR(ComponentVersion.componentName)),
			FString(UTF8_TO_TCHAR(ComponentVersion.versionIdentifier)),
			FString(UTF8_TO_TCHAR(ComponentVersion.buildIdentifier)),
			FString(UTF8_TO_TCHAR(ComponentVersion.buildDateTime)),
			FString(UTF8_TO_TCHAR(ComponentVersion.sourceIdentifier))
		};
		logCV.Append("\nComponent '").Append(*SpacesComponentVersion.Name).Append("'\n     Version Identifier: ").Append(*SpacesComponentVersion.VersionId).Append("\n     Build Identifier: ").Append(*SpacesComponentVersion.BuildId).Append("\n     Build Date Time: ").Append( *SpacesComponentVersion.BuildDateTime);
		ComponentVersions.Add(MoveTemp(SpacesComponentVersion));
	}
	UE_LOG(LogSpacesRuntime, Log, TEXT("%s"), *logCV);
	
	return true;
}

void FSpacesCoreExtension::RuntimeCheck()
{
	const bool bCheckInstalledRuntime = READ_CONFIG_VALUE(SpacesRuntime, SpacesRuntimeSettings, bCheckInstalledRuntime, Bool, true);
	if(!bCheckInstalledRuntime)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Skip runtime check detected, skipping check."));
		return;
	}

#if PLATFORM_ANDROID
    UE_LOG(LogSpacesRuntime, Log, TEXT("Checking for the Snapdragon Spaces Services runtime."));

    FRuntimeCheckerWrapper RuntimeChecker;

    if(!RuntimeChecker.CheckInstalled())
    {
        FText DialogHeader = LOCTEXT("DialogHeader","Services not installed");
        FText DialogContent = LOCTEXT("DialogContent","Please install the Snapdragon Spaces Services before starting this application!");

        FMessageDialog::Open(
            EAppMsgType::Ok,
            DialogContent,
            &DialogHeader
            );
        UE_LOG(LogSpacesRuntime, Error, TEXT("Snapdragon Spaces Services is not installed! Application will be closed."));
        FGenericPlatformMisc::RequestExit(false);
    }

#else  //PLATFORM_ANDROID
    UE_LOG(LogSpacesRuntime, Log, TEXT("Runtime check not done, invalid platform."));
#endif
}

FSpacesCoreExtension::FSpacesCoreExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Startup SpacesBaseExtension"));
	RegisterOpenXRExtensionModularFeature();
    RuntimeCheck();

	CaptureSupport.Add(EARCaptureType::SceneUnderstanding, MakeUnique<FSpacesSceneUnderstanding>());
	CaptureSupport.Add(EARCaptureType::SpatialMapping, MakeUnique<FSpacesSpatialMapping>());
	CaptureSupport.Add(EARCaptureType::Camera, MakeUnique<FSpacesCameraCapture>());
	CaptureSupport.Add(EARCaptureType::HandMesh, MakeUnique<FSpacesHandMeshCapture>());
	
	EKeys::AddMenuCategoryDisplayInfo("SnapdragonSpaces", LOCTEXT("SnapdragonSpacesSubCategory", "Snapdragon Spaces"), TEXT("GraphEditor.PadEvent_16x"));
}

FSpacesCoreExtension::~FSpacesCoreExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Shutdown SpacesBaseExtension"));
#if PLATFORM_ANDROID
	if (LoaderHandle)
	{
		FPlatformProcess::FreeDllHandle(LoaderHandle);
		LoaderHandle = nullptr;
	}
#endif
	CaptureSupport.Empty();
	UnregisterOpenXRExtensionModularFeature();
}

bool FSpacesCoreExtension::GetCustomLoader(PFN_xrGetInstanceProcAddr* OutGetProcAddr)
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Spaces GetCustomLoader"));
#if PLATFORM_ANDROID
	// clear errors
	dlerror();

	LoaderHandle = FPlatformProcess::GetDllHandle(TEXT("libopenxr_loader.so"));
	if (LoaderHandle == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Unable to load libopenxr_loader.so, error %s"), ANSI_TO_TCHAR(dlerror()));
		return false;
	}

	// clear errors
	dlerror();

	SpacesOpenXRIntercept::xrGetInstanceProcAddr_Native = (PFN_xrGetInstanceProcAddr)FPlatformProcess::GetDllExport(LoaderHandle, TEXT("xrGetInstanceProcAddr"));
	if (SpacesOpenXRIntercept::xrGetInstanceProcAddr_Native == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Unable to load OpenXR xrGetInstanceProcAddr, error %s"), ANSI_TO_TCHAR(dlerror()));
		return false;
	}
	*OutGetProcAddr = &SpacesOpenXRIntercept::xrGetInstanceProcAddr_Intercept;
	
	extern struct android_app* GNativeAndroidApp;
	PFN_xrInitializeLoaderKHR xrInitializeLoaderKHR;
	SpacesOpenXRIntercept::xrGetInstanceProcAddr_Intercept(XR_NULL_HANDLE, "xrInitializeLoaderKHR", (PFN_xrVoidFunction*)&xrInitializeLoaderKHR);
	if (xrInitializeLoaderKHR == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Unable to load OpenXR xrInitializeLoaderKHR"));
		return false;
	}
	XrLoaderInitInfoAndroidKHR LoaderInitializeInfoAndroid;
	LoaderInitializeInfoAndroid.type = XR_TYPE_LOADER_INIT_INFO_ANDROID_KHR;
	LoaderInitializeInfoAndroid.next = NULL;
	LoaderInitializeInfoAndroid.applicationVM = GNativeAndroidApp->activity->vm;
	LoaderInitializeInfoAndroid.applicationContext = GNativeAndroidApp->activity->clazz;
	if(XR_FAILED(xrInitializeLoaderKHR((XrLoaderInitInfoBaseHeaderKHR*)&LoaderInitializeInfoAndroid)))
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to run xrInitializeLoaderKHR"));
		FPlatformProcess::FreeDllHandle(LoaderHandle);
		LoaderHandle = nullptr;
		return false;
	}

    UE_LOG(LogSpacesRuntime, Log, TEXT("Loaded Snapdragon Spaces OpenXR Loader"));

	return true;
#else //PLATFORM_ANDROID
	UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces GetCustomLoader platform not supported."));
	return false;
#endif //PLATFORM_ANDROID

}

void FSpacesCoreExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
	SPACES_OPENXR_FUNCTION_TYPES(GET_XR_ENTRYPOINT)
}

//TODO(AS): Stereo only device for testing purpose, will change as we will need to use the phone screen independently
bool FSpacesCoreExtension::IsStandaloneStereoOnlyDevice()
{
	return true;
}

bool FSpacesCoreExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	OutExtensions.Add("XR_KHR_convert_timespec_time");
	OutExtensions.Add(XR_QCOM_COMPONENT_VERSIONING_EXTENSION_NAME);
	// OutExtensions.Add("XR_KHR_loader_init_android");
	return true;
}

IOpenXRCustomCaptureSupport* FSpacesCoreExtension::GetCustomCaptureSupport(const EARCaptureType CaptureType)
{
	const auto CaptureSupportEntry = CaptureSupport.Find(CaptureType);
	if(CaptureSupportEntry == nullptr) return nullptr;
	CaptureSupportEntry->Get()->Initialize();
	return CaptureSupportEntry->Get();
}

//TODO(AS): Will need to implement an spectator screen for the phone display
bool FSpacesCoreExtension::GetSpectatorScreenController(FHeadMountedDisplayBase* InHMDBase,
	TUniquePtr<FDefaultSpectatorScreenController>& DefaultSpectatorScreenController)
{
#if PLATFORM_ANDROID
	DefaultSpectatorScreenController = nullptr;
	return true;
#else // PLATFORM_ANDROID
	DefaultSpectatorScreenController = MakeUnique<FDefaultSpectatorScreenController>(InHMDBase);
	return false;
#endif // PLATFORM_ANDROID
}

const void* FSpacesCoreExtension::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Spaces OnCreateSession"));
#if PLATFORM_ANDROID
	if (GRHISupportsRHIThread && GIsThreadedRendering && GUseRHIThread_InternalUseOnly)
	{
		//TODO(AS): Investigate why QtiOpenXRHMD plugin is disabling RHI thread
	    //TODO(AS): It's the same implementation as the Oculus OpenXR plugin
		SetRHIThreadEnabled(false, false);
	}
#endif // PLATFORM_ANDROID
	return InNext;
}

const void* FSpacesCoreExtension::OnEndProjectionLayer(XrSession InSession, int32 InLayerIndex, const void* InNext,
	XrCompositionLayerFlags& OutFlags)
{
	UE_LOG(LogSpacesRuntime, Verbose, TEXT("Spaces OnEndProjectionLayer"));
	OutFlags |= XR_COMPOSITION_LAYER_CORRECT_CHROMATIC_ABERRATION_BIT;
	OutFlags |= XR_COMPOSITION_LAYER_BLEND_TEXTURE_SOURCE_ALPHA_BIT;
	return InNext;
}

FString FSpacesCoreExtension::GetDisplayName()
{
	return FString(TEXT("Spaces Base Runtime"));
}

#undef LOCTEXT_NAMESPACE