/******************************************************************************
 * File: SpacesExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesExtension.h"
#include "OpenXRHMD/Private/OpenXRHMD.h"
#include "SpacesSessionConfig.h"
#include "JNI/RuntimeCheckerWrapper.h"


void FSpacesExtension::SetSessionHandle(const XrSession& InSession)
{
	SessionHandle = InSession;
}

void FSpacesExtension::SetInstanceHandle(const XrInstance& InInstance)
{
	InstanceHandle = InInstance;
}

bool FSpacesExtension::IsValidExtension(FSpacesExtension* Extension)
{
#if PLATFORM_ANDROID
	TArray<FString> Permissions;
	Extension->GetRequiredPermissions(Permissions);
	
	FPermissionHelperWrapper PermissionHelper;
	for(auto Permission : Permissions)
	{
		if(!PermissionHelper.CheckPermission(Permission))
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Cannot load %s, missing required permission %s."), *(Extension->GetDisplayName()), *Permission);
			return false;
		}
	}
	
	if(Extension->IsRuntimeCameraPermissionsRequired())
	{
		FRuntimeCheckerWrapper RuntimeChecker;
		if(!RuntimeChecker.CheckCameraPermissions())
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Cannot load %s, missing required runtime camera permission."), *(Extension->GetDisplayName()));
			return false;	
		}
	}
#endif
	return true;
}

XrSession FSpacesExtension::GetSessionHandle() const
{
	return SessionHandle;
}

XrInstance FSpacesExtension::GetInstanceHandle() const
{
	return InstanceHandle;
}

FOpenXRHMD* FSpacesExtension::GetXRSystem() const
{
	const auto XRSystem = static_cast<FOpenXRHMD*>(GEngine->XRSystem.Get());
	if(XRSystem == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Invalid FOpenXRHM, cannot get XRSystem"));
	}
	return XRSystem;
}

bool FSpacesCaptureExtension::Reset()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Reset"));
	TArray<EARCaptureType> CurrentEnabledTypes = EnabledTypes;
	bool AllStopped = true;
	for(const EARCaptureType Type : CurrentEnabledTypes)
	{
		AllStopped = DisableCaptureType(Type) && AllStopped;
	}
	if(!AllStopped){ return false; }
	bool AllStarted = true;
	for(const EARCaptureType Type : CurrentEnabledTypes)
	{
		AllStarted = EnableCaptureType(Type) && AllStarted;
	}
	if(!AllStarted){ return false; }
	return true;
}

bool FTrackedPlaneHolder::IsYZPlane() const
{
	if(!FMath::IsNearlyEqual(Extents.Z, 0.0f)){ return true;}
	return false;
}

void FTrackedPlaneHolder::TransformToXYPlane()
{
	if(!IsYZPlane()){ return; }

	bool ReverseOrientation = false;
	if(Vertices.Num() > 3)
	{
		const FVector P1 = Vertices[Indices[0]];
		const FVector P2 = Vertices[Indices[1]];
		const FVector P3 = Vertices[Indices[2]];
		const FVector V1 = (P1 - P3).GetSafeNormal();
		const FVector V2 = (P2 - P3).GetSafeNormal();
		ReverseOrientation = FVector::CrossProduct(V1, V2).X < 0;
	}
	const FTransform Rotation(FQuat(FVector::RightVector, FMath::DegreesToRadians(ReverseOrientation? 90 :-90)).Rotator());
	for(FVector& Vertex: Vertices)
	{
		Vertex = Rotation.TransformVector(FVector(Vertex.X,Vertex.Y,Vertex.Z));
		//Vertex = FVector3f(Vertex_.X, Vertex_.Y, Vertex_.Z);
	}
	LocalToTrackingTransform = Rotation.Inverse() * LocalToTrackingTransform;
	Extents = Rotation.TransformVector(Extents).GetAbs();
	
}

void FSpacesCaptureExtension::OnStartARSession(UARSessionConfig* SessionConfig)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnStartARSession"));
	if(SessionConfig == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("SessionConfig is NULL"));
		return;
	}
	USpacesSessionConfig* NewSessionConfig = Cast<USpacesSessionConfig>(SessionConfig);
	if(NewSessionConfig == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("SessionConfig is not a SpacesSessionConfig!"));
		OnStopARSession();
		return;
	}
	const bool bNeedsReset = CurrentSessionConfig != NewSessionConfig;
	CurrentSessionConfig = NewSessionConfig;
	if(bARSessionPaused && !bNeedsReset){ OnResume(); }
	bARSessionPaused = false;
	if(bNeedsReset){ Reset();}
}

void FSpacesCaptureExtension::OnPauseARSession()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnPauseARSession"));
    bARSessionPaused = true;
	OnPause();
}

void FSpacesCaptureExtension::OnStopARSession()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnStopARSession"));
	TArray<EARCaptureType> CurrentEnabledTypes = EnabledTypes;
	bARSessionPaused = false;
	for(const EARCaptureType Type : CurrentEnabledTypes)
	{
		DisableCaptureType(Type);
	}
}

bool FSpacesCaptureExtension::EnableCaptureType(EARCaptureType InType)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("EnableCaptureType"));
	
	if(!CanEnableExtension())
	{
		/* Cannot enable this extension */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Cannot enable this extension"));
		return false;
	}
	
	if(bARSessionPaused)
	{
		/* AR Session is not running */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("AR Session is not running"));
		return false;
	}
	
	TArray<EARCaptureType> SupportedTypes;
	if(!GetSupportedCaptureTypes(SupportedTypes))
	{
		/* Failed to get supported types */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to get supported types"));
		return false;
	}
	
	if(!SupportedTypes.Contains(InType))
	{
		/* Invalid capture type */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Invalid capture type"));
		return false;
	}
	if(EnabledTypes.Contains(InType))
	{
		/* Already enabled */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Already enabled"));
		return false;
	}
	EnabledTypes.Add(InType);
	if(!OnEnabledCaptureType(InType))
	{
		/* Failed to enable*/
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to enable"));
		EnabledTypes.Remove(InType);
		return false;
	}
	if(EnabledTypes.Num() > 1)
	{
		/* There's more than 1 type enabled, feature is already started */
		UE_LOG(LogSpacesRuntime, Log, TEXT("There's more than 1 type enabled, feature is already started"));
		return true;
	}
	/* First capture type enabled, the feature must start */
	if(!OnStart())
	{
		/* Failed to start */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to start"));
		EnabledTypes.Remove(InType);
		return false;
	}
	return true;
}

bool FSpacesCaptureExtension::DisableCaptureType(EARCaptureType InType)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("DisableCaptureType"));
	
	if(!EnabledTypes.Contains(InType))
	{
		/* Capture type is not enabled, cannot disable */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Capture type is not enabled, cannot disable"))
		return false;
	}
	EnabledTypes.Remove(InType);
	if(!OnDisabledCaptureType(InType))
	{
		/* Failed to disable capture type */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to disable capture type"))
		EnabledTypes.Add(InType);
		return false;
	}
	if(EnabledTypes.Num() > 0)
	{
		/* There are still capture types enabled, feature is must keep running */
		UE_LOG(LogSpacesRuntime, Log, TEXT("There are still capture types enabled, feature is must keep running"))
		return true;
	}
	/* Last capture type disabled, the feature must stop */
	if(!OnStop())
	{
		/* Failed to stop */
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed to stop"))
		EnabledTypes.Add(InType);
		return false;
	}
	return true;
}

bool FSpacesCaptureExtension::OnToggleARCapture(const bool bOnOff)
{
	/* This shouldn't be called */
	return false;
}

bool FSpacesCaptureExtension::IsEnabled() const
{
	return EnabledTypes.Num() > 0;
}

const TArray<EARCaptureType>& FSpacesCaptureExtension::GetEnabledCaptureTypes()
{
	return EnabledTypes;
}
