/******************************************************************************
 * File: SpacesExtension.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#pragma once

#include "CoreMinimal.h"
#include "External/openxr/openxr.h"
#include "SpacesCommon.h"
#include "IOpenXRExtensionPlugin.h"
#include "MRMeshBufferDefines.h"

#if PLATFORM_ANDROID
#include "JNI/PermissionHelperWrapper.h"
#endif

#include "SpacesXrFunction.h"
#include "SpacesXrHandle.h"

#include "SpacesSessionConfig.h"

class FOpenXRHMD;

/***
 * @brief List of Spaces implemented extensions
 * @details Adding a new entry to this list autogenerate the necessary glue code
 * @details Before adding a new entry, the extension class must be created following
 * this naming rule, FSpaces<NAME>Extension, where <NAME> is the name of the extension.
 */
#define SPACES_EXTENSION_LIST(_)	\
_(SpatialAnchor)					\
_(PlaneDetection)					\
_(ImageTracking)					\
_(HandTracking)						\
_(HitTesting)						\
_(Meshing)							\
_(CameraAccess)

#define SPACES_REQUIRED_EXTENSION_LIST(_)	\
_(Core)										\
_(InputXRController)						\
_(InputHostController)						\


#define SPACES_EXTENSION_ENUM_ENTRY(Name) Name,
/***
 * @brief Spaces implemented extensions enum
 */
enum class ESpacesExtension : uint8
{
	SPACES_EXTENSION_LIST(SPACES_EXTENSION_ENUM_ENTRY)
	SPACES_REQUIRED_EXTENSION_LIST(SPACES_EXTENSION_ENUM_ENTRY)
	Max
};
#undef  SPACES_EXTENSION_ENUM_ENTRY

#define SPACES_EXTENSION_ENUMFLAGS_ENTRY(Name) Name = 1<<((uint8)ESpacesExtension::Name),
#define SPACES_EXTENSION_ENUMFLAGS_REQUIRED_VALUE(Name) Name +
/***
 * @brief Spaces implemented extensions flags
 */
enum class ESpacesExtensionFlags : uint64
{
	None					= 0	,
			
	SPACES_EXTENSION_LIST(SPACES_EXTENSION_ENUMFLAGS_ENTRY)		
	SPACES_REQUIRED_EXTENSION_LIST(SPACES_EXTENSION_ENUMFLAGS_ENTRY)

	Max						= 1<<((uint8)ESpacesExtension::Max),

	All						= Max-1,
	Required				= SPACES_REQUIRED_EXTENSION_LIST(SPACES_EXTENSION_ENUMFLAGS_REQUIRED_VALUE) 0, //0 is added for closing the macro sum
	Optional				= All & ~Required
	
};
#undef SPACES_EXTENSION_ENUMFLAGS_REQUIRED_VALUE
#undef SPACES_EXTENSION_ENUMFLAGS_ENTRY


#define SPACES_EXTENSION_CASE_STR(name) case ESpacesExtension::name: return TEXT(#name);
namespace SnapdragonSpaces
{
	constexpr const TCHAR* SpacesExtensionToStr(ESpacesExtension e)
	{
		switch (e) {										
			SPACES_EXTENSION_LIST(SPACES_EXTENSION_CASE_STR)	
			SPACES_REQUIRED_EXTENSION_LIST(SPACES_EXTENSION_CASE_STR)	
			default: return TEXT("Unknown");				
		}	
	}
}
#undef SPACES_EXTENSION_CASE_STR

/***
 * @brief Spaces extension base class
 * @details Base class for all Spaces extensions.
 */

class FSpacesExtension : public IOpenXRExtensionPlugin
{
public:

	/***
	 * @brief Instantiation helper method
	 * @details Method used to create each extension instance.
	 * Only valid to specialize using derived FSpacesExtension classes.
	 * @param Extension Extension class to instantiate
	 * @return Unique pointer with the extension instance, or invalid if instantiation failed.
	 */
	template <typename Extension>
	static TUniquePtr<FSpacesExtension> TryCreateExtension()
	{
		FSpacesExtension* NewExtension = new Extension();
		if(!IsValidExtension(NewExtension))
		{
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Error loading extension: %s"), *(NewExtension->GetDisplayName()));
			delete NewExtension;
			return TUniquePtr<FSpacesExtension>();
		}
		UE_LOG(LogSpacesRuntime, Log, TEXT("Loaded extension: %s"), *(NewExtension->GetDisplayName()));
		return TUniquePtr<FSpacesExtension>(NewExtension);
	}

	virtual ~FSpacesExtension() override = default;

	XrSession GetSessionHandle() const;
	XrInstance GetInstanceHandle() const;
	FOpenXRHMD* GetXRSystem() const;
	
protected:

	void SetSessionHandle(const XrSession& InSession);
	void SetInstanceHandle(const XrInstance& InInstance);

#if PLATFORM_ANDROID
	virtual void GetRequiredPermissions(TArray<FString>& Permissions) const {};
	virtual bool IsRuntimeCameraPermissionsRequired() const {return false;};
#endif
	
private:

	static bool IsValidExtension(FSpacesExtension* Extension);
	
	XrSession SessionHandle = XR_NULL_HANDLE;
	XrInstance InstanceHandle = XR_NULL_HANDLE;
};

/***
 * @brief Spaces capture extension base class
 * @details Base class for AR capable Spaces extensions.
 */
class FSpacesCaptureExtension : public FSpacesExtension, public IOpenXRCustomCaptureSupport
{
public:
	
	virtual void OnStartARSession(UARSessionConfig* SessionConfig) override final;
	virtual void OnPauseARSession() override final;
	virtual void OnStopARSession() override final;

	bool EnableCaptureType(EARCaptureType InType);
	bool DisableCaptureType(EARCaptureType InType);
	virtual bool GetSupportedCaptureTypes(TArray<EARCaptureType>& OutTypes) = 0;
	
	virtual bool OnToggleARCapture(const bool bOnOff) override final;
	virtual bool IsEnabled() const override final;

protected:

	FSpacesCaptureExtension() = default;
	
	bool bARSessionPaused = false;
	USpacesSessionConfig*  CurrentSessionConfig = nullptr;

	const TArray<EARCaptureType>& GetEnabledCaptureTypes();
	
	virtual bool CanEnableExtension() const = 0;

	virtual bool OnEnabledCaptureType(EARCaptureType InType) = 0;
	virtual bool OnDisabledCaptureType(EARCaptureType InType) = 0;

	virtual bool OnStart() = 0;
	virtual void OnPause() = 0;
	virtual void OnResume() = 0;
	virtual bool OnStop() = 0;
	
private:
	
	bool Reset();

	TArray<EARCaptureType> EnabledTypes;
	
};

struct FTrackedHolder
{
	FGuid Id;
	FGuid ParentId;
	FTransform LocalToTrackingTransform;
	EARTrackingState TrackingState = EARTrackingState::Unknown;
	EARObjectClassification Classification = EARObjectClassification::NotApplicable;
	EARSpatialMeshUsageFlags SpatialMeshUsageFlags = EARSpatialMeshUsageFlags::NotApplicable;

};

struct FTrackedMeshHolder : FTrackedHolder
{
	TArray<FVector> Vertices;
	TArray<FVector> Normals;
	TArray<MRMESH_INDEX_TYPE> Indices;
};

struct FTrackedPlaneHolder : FTrackedMeshHolder
{
	FVector Extents = FVector::ZeroVector;
	EARPlaneOrientation Orientation = EARPlaneOrientation::Horizontal;

	bool IsYZPlane() const;
	void TransformToXYPlane();
};