/******************************************************************************
 * File: SpacesHGesturesBlueprintLibrary.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesHGesturesBlueprintLibrary.h"

#include "SpacesExtension.h"
#include "HandTracking/SpacesHandTrackingExtension.h"
#include "SpacesRuntimeModule.h"
#include "Materials/MaterialInterface.h"

#include "ProceduralMeshComponent.h"

FSpacesHandGestureData USpacesHGesturesBlueprintLibrary::GetXRHandGestureData(ESpacesHandType HandType)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::HandTracking))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("HandTracking extension disabled."));
		return {ESpacesHandGestureType::ERROR,0.0,0.0};
	}

	const FSpacesHandTrackingExtension* HaTExtension = static_cast<const FSpacesHandTrackingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::HandTracking].Get());
	if(HaTExtension == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("HandTracking extension disabled."));
		return {ESpacesHandGestureType::ERROR,0.0,0.0};
	}
	
	const XrHandGestureQCOM HandGestureRes = HaTExtension->GetHandGestureData((int)HandType);
	
	int GestureTypeInt = HandGestureRes.gesture;
	if(GestureTypeInt == -1)
	{
		
		GestureTypeInt = 22;
		
	}
	
	const ESpacesHandGestureType GestureType = static_cast<ESpacesHandGestureType>(GestureTypeInt);
	return  {GestureType, HandGestureRes.gestureRatio, HandGestureRes.flipRatio};
}

bool USpacesHGesturesBlueprintLibrary::IsHandTracked(ESpacesHandType HandType)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::HandTracking))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return false;
	}

	const FSpacesHandTrackingExtension* HaTExtension = static_cast<const FSpacesHandTrackingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::HandTracking].Get());
	if(HaTExtension == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return false;
	}
	
	return HaTExtension->HandStateJointPosesAvailable((int)HandType);
}

void USpacesHGesturesBlueprintLibrary::SetHandMeshStatus(bool Active)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::HandTracking))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return;
	}

	FSpacesHandTrackingExtension* HaTExtension = static_cast<FSpacesHandTrackingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::HandTracking].Get());
	
	HaTExtension->SetHandMeshingStatus(Active);
}

void USpacesHGesturesBlueprintLibrary::SetHandTrackingState(bool Active)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::HandTracking))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return;
	}
	
	FSpacesHandTrackingExtension* HaTExtension = static_cast<FSpacesHandTrackingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::HandTracking].Get());
	if(HaTExtension == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return;
	}
	HaTExtension->SetHandTrackingState(Active);
}

bool USpacesHGesturesBlueprintLibrary::GetHandMeshStatus()
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::HandTracking))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return false;
	}

	const FSpacesHandTrackingExtension* HaTExtension = static_cast<const FSpacesHandTrackingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::HandTracking].Get());
	if(HaTExtension == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return false;
	}
	
	return HaTExtension->GetHandMeshingStatus();
}

void USpacesHGesturesBlueprintLibrary::RenderHandMesh(ESpacesHandType HandIndex, UMaterialInterface* Material, AActor*& HandMeshActor, int& OriginalVerticesNum)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::HandTracking))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Hand Tracking extension disabled."));
		return;
	}
	
	const FSpacesHandTrackingExtension* HaTExtension = static_cast<const FSpacesHandTrackingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::HandTracking].Get());
	
	TArray<FVector> Vertices;
	TArray<FVector> Normals;
	TArray<int32> Indices;
	FTransform HandMeshTransform;
	
	if(!HaTExtension->GetHandMeshData((EControllerHand)HandIndex, Vertices, Normals, Indices, HandMeshTransform))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("GetHandMesh data failed."));
		return;
	}
	
	UProceduralMeshComponent* ProceduralHandMeshComponent = nullptr;
	
	if(HandMeshActor == nullptr)
	{
		FString HandMeshActorName = "HandMesh_" + FString((int)HandIndex == 0 ? "Left" : "Right");
		HandMeshActor = NewObject<AActor>(GWorld->GetCurrentLevel(), FName(HandMeshActorName));
		ProceduralHandMeshComponent = NewObject<UProceduralMeshComponent>(HandMeshActor, UProceduralMeshComponent::StaticClass());
		ProceduralHandMeshComponent->RegisterComponentWithWorld(GWorld);
		ProceduralHandMeshComponent->SetCollisionProfileName(UCollisionProfile::NoCollision_ProfileName);
		OriginalVerticesNum = Vertices.Num();
		ProceduralHandMeshComponent->CreateMeshSection_LinearColor(0, Vertices, Indices, Normals, {}, {}, {}, false);
		ProceduralHandMeshComponent->SetMaterial(0, Material);
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Creating handmesh actor:%s"),*HandMeshActorName);
	}
	else
	{
		ProceduralHandMeshComponent = Cast<UProceduralMeshComponent>(HandMeshActor->GetComponentByClass(UProceduralMeshComponent::StaticClass()));
		if (ProceduralHandMeshComponent)
		{
			if(Vertices.Num() != OriginalVerticesNum)
			{
				//Recreate mesh section if current vertices amount is different from the original vertices amount (amount of vertices when section is created)
				ProceduralHandMeshComponent->ClearMeshSection(0);
				ProceduralHandMeshComponent->CreateMeshSection_LinearColor(0, Vertices, Indices, Normals, {}, {}, {}, false);
				OriginalVerticesNum = Vertices.Num();
				UE_LOG(LogSpacesRuntime, Warning, TEXT("Recreating handmesh section 0, vertices num: %d"),OriginalVerticesNum);
				return;
			}
			else
			{
				ProceduralHandMeshComponent->UpdateMeshSection_LinearColor(0, Vertices, Normals, {}, {}, {});
			}
		}
	}

	if(ProceduralHandMeshComponent)
	{
		ProceduralHandMeshComponent->SetWorldTransform(HandMeshTransform);
	}
}

