﻿/******************************************************************************
 * File: SpacesOpenXRIntercept.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */


#include "SpacesOpenXRIntercept.h"
#include "SpacesCommon.h"
#include "Engine.h"
#include "SubSystems/SpacesEngineSubsystem.h"
#include "SpacesRuntimeModule.h"
#include "Input/SpacesInputHostControllerExtension.h"
#include "Runtime/Launch/Resources/Version.h"


/************************************************************************************************************/

#define INTERCEPT_CALL_DEFINITION(Function)							\
PFN_##Function SpacesOpenXRIntercept::Function##_Native = nullptr;	\
XrResult SpacesOpenXRIntercept::Function##_Intercept

#define INTERCEPT_CALL_FUNCTION_ITEM(Function)								\
if(strcmp(#Function, name) == 0){											\
	XrResult Result = xrGetInstanceProcAddr_Native(							\
		instance,															\
		name,																\
		reinterpret_cast<PFN_xrVoidFunction*>(&Function##_Native)			\
	);																		\
	*function = reinterpret_cast<PFN_xrVoidFunction>(&Function##_Intercept);\
	return Result;															\
}


INTERCEPT_CALL_DEFINITION(xrGetInstanceProcAddr)(
	XrInstance                                  instance,
	const char*                                 name,
	PFN_xrVoidFunction*                         function)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Intercepting xrGetInstanceProcAddr, requested function: %hs"), name);

#pragma warning( push )
	//Disable reinterpret cast function pointer warnings
	//https://docs.microsoft.com/en-us/cpp/error-messages/compiler-warnings/compiler-warning-level-3-c4191?view=msvc-170
	#pragma warning( disable : 4191)
	INTERCEPT_CALL_LIST(INTERCEPT_CALL_FUNCTION_ITEM)
#pragma warning( pop )
	
	return xrGetInstanceProcAddr_Native(instance, name, function);
}

/************************************************************************************************************/



/************************************ Intercept functions definition ****************************************/

INTERCEPT_CALL_DEFINITION(xrCreateReferenceSpace)(
	XrSession                                   session,
	const XrReferenceSpaceCreateInfo*           createInfo,
	XrSpace*                                    space)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Intercepted xrCreateReferenceSpace call"));
	UE_LOG(LogSpacesRuntime, Log, TEXT("xrCreateReferenceSpace_Native = %p"), xrCreateReferenceSpace_Native);
	return xrCreateReferenceSpace_Native(session, createInfo, space);
}

bool SpacesOpenXRIntercept::bPassthroughSupported = false;

INTERCEPT_CALL_DEFINITION(xrEnumerateEnvironmentBlendModes)(
	XrInstance                                  instance,
	XrSystemId                                  systemId,
	XrViewConfigurationType                     viewConfigurationType,
	uint32_t                                    environmentBlendModeCapacityInput,
	uint32_t*                                   environmentBlendModeCountOutput,
	XrEnvironmentBlendMode*                     environmentBlendModes)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Intercepted xrEnumerateEnvironmentBlendModes call"));
	UE_LOG(LogSpacesRuntime, Log, TEXT("xrEnumerateEnvironmentBlendModes_Native = %p"), xrEnumerateEnvironmentBlendModes_Native);

	//Getting subsystem
	const XrResult Result = xrEnumerateEnvironmentBlendModes_Native(instance, systemId, viewConfigurationType, environmentBlendModeCapacityInput, environmentBlendModeCountOutput, environmentBlendModes);
	const auto SpacesEngineSubsystem = GEngine->GetEngineSubsystem<USpacesEngineSubsystem>();

	//Check valid device for Passthrough
	for(uint32_t i = 0; i<environmentBlendModeCapacityInput; ++i)
	{
		if(environmentBlendModes[i] == XR_ENVIRONMENT_BLEND_MODE_ALPHA_BLEND)
		{
			bPassthroughSupported =  true;
			break;
		}
	}

	//Forcing Passthrough if enabled from init
	if (SpacesEngineSubsystem->GetPassthroughEnabled() && bPassthroughSupported)
	{
		for(uint32_t i = 0; i<environmentBlendModeCapacityInput; ++i)
		{
			UE_LOG(LogSpacesRuntime, Log, TEXT("Got BlendMode %s, changed by XR_ENVIRONMENT_BLEND_MODE_ALPHA_BLEND"), SnapdragonSpaces::XrEnumStr(environmentBlendModes[i]));
			environmentBlendModes[i] = XR_ENVIRONMENT_BLEND_MODE_ALPHA_BLEND;
		}
	}

	return Result;
}

INTERCEPT_CALL_DEFINITION(xrEndFrame)(
    XrSession                                   session,
    const XrFrameEndInfo*                       frameEndInfo)
{
	//Copy const Info and modify it on interception
	const auto SpacesEngineSubsystem = GEngine->GetEngineSubsystem<USpacesEngineSubsystem>();
	XrFrameEndInfo frameEndInfoCopy =  *frameEndInfo;
	frameEndInfoCopy.environmentBlendMode = SpacesEngineSubsystem->GetPassthroughEnabled() ? XR_ENVIRONMENT_BLEND_MODE_ALPHA_BLEND : XR_ENVIRONMENT_BLEND_MODE_OPAQUE;

#if ENGINE_MAJOR_VERSION == 5
	// TODO(GG): this code is to remove a workaround included in UE5 for a specific issue (UE-145546)
	for(uint32_t n = 0; n < frameEndInfoCopy.layerCount; n ++)
	{
		if(frameEndInfoCopy.layers[n]->type == XR_TYPE_COMPOSITION_LAYER_PROJECTION)
		{
			reinterpret_cast<XrCompositionLayerProjection*>(const_cast<XrCompositionLayerBaseHeader*>(frameEndInfoCopy.layers[n]))->layerFlags &= ~XR_COMPOSITION_LAYER_UNPREMULTIPLIED_ALPHA_BIT;
		}
	}
#endif
	
	return xrEndFrame_Native(session,&frameEndInfoCopy);
}

INTERCEPT_CALL_DEFINITION(xrApplyHapticFeedback)(
	XrSession									session,
	const XrHapticActionInfo*					hapticActionInfo,
	const XrHapticBaseHeader*					hapticFeedback)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Intercepted xrApplyHapticFeedback call"));
	UE_LOG(LogSpacesRuntime, Log, TEXT("xrApplyHapticFeedback_Native = %p"), xrApplyHapticFeedback_Native);
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::InputHostController))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces InputHostController extension disabled."));
		return xrApplyHapticFeedback_Native(session, hapticActionInfo, hapticFeedback);
	}
	
	const FSpacesInputHostControllerExtension* InputHostControllerExtension = static_cast<const FSpacesInputHostControllerExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::InputHostController].Get());

	if(InputHostControllerExtension == nullptr)
	{
		return xrApplyHapticFeedback_Native(session, hapticActionInfo, hapticFeedback);
	}
	
	const XrDuration HapticsDuration = (InputHostControllerExtension->GetHapticsDuration()*1000000000.0f);
	XrHapticVibration HapticValue;
	HapticValue.type = XR_TYPE_HAPTIC_VIBRATION;
	HapticValue.next = nullptr;
	HapticValue.duration = HapticsDuration;
	HapticValue.frequency = InputHostControllerExtension->GetHapticsFrequency();
	HapticValue.amplitude = InputHostControllerExtension->GetHapticsAmplitude();
	
	return xrApplyHapticFeedback_Native(session, hapticActionInfo, (const XrHapticBaseHeader*)&HapticValue);
}