﻿/******************************************************************************
 * File: SpacesRuntimeBlueprintLibrary.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesRuntimeBlueprintLibrary.h"

#include "Engine/Engine.h"
#include "SpatialAnchor/SpacesSpatialAnchor.h"
#include "SpatialAnchor/SpacesSpatialAnchorExtension.h"
#include "ARPin.h"
#include "IXRTrackingSystem.h"
#include "SpacesCoreExtension.h"
#include "SpacesRuntimeModule.h"
#include "Meshing/SpacesMeshingExtension.h"
#include "ProceduralMeshComponent.h"
#include "Async/Async.h"
#include "SubSystems/SpacesEngineSubsystem.h"
#include "Misc/DateTime.h"
#include "Async/Async.h"
#include "SpacesRuntimeSettings.h"
#include "JNI/SpacesControllerWrapper.h"
#include "GameFramework/PlayerController.h"
#include "InputCoreTypes.h"
#include "Input/SpacesInputHostControllerExtension.h"
#include "CameraAccess/SpacesCameraAccessExtension.h"
#include "JNI/SpacesSettingsWrapper.h"

bool USpacesRuntimeBlueprintLibrary::GetAnchorInfoFromARPin(UARPin* Pin, int64& AnchorHandle, int64& AnchorSpaceHandle)
{
	if(!IsValid(Pin) || Pin->GetNativeResource() == nullptr) return false;

	const FSpacesSpatialAnchor* Anchor = static_cast<FSpacesSpatialAnchor*>(Pin->GetNativeResource());

	if(Anchor == nullptr) return false;

	AnchorHandle = (int64)Anchor->GetAnchor();
	AnchorSpaceHandle = (int64)Anchor->GetAnchorSpace();

	return true;
	
}

bool USpacesRuntimeBlueprintLibrary::LoadSavedAnchor(FName InName)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::SpatialAnchor))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spatial Anchor extension disabled."));
		return false;
	}

	const FSpacesSpatialAnchorExtension* SpacesSpatialAnchor = static_cast<const FSpacesSpatialAnchorExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::SpatialAnchor].Get());
	if(SpacesSpatialAnchor == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spatial Anchor extension disabled."));
		return false;
	}
	
	return false;
}

TArray<FName> USpacesRuntimeBlueprintLibrary::GetSavedAnchorNames()
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::SpatialAnchor))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spatial Anchor extension disabled."));
		return TArray<FName>();
	}

	const FSpacesSpatialAnchorExtension* SpacesSpatialAnchor = static_cast<const FSpacesSpatialAnchorExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::SpatialAnchor].Get());
	if(SpacesSpatialAnchor == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spatial Anchor extension disabled."));
		return TArray<FName>();
	}
	
	TArray<XrSpatialAnchorPersistenceNameMSFT> AnchorNames = SpacesSpatialAnchor->GetSavedAnchorNames();
	TArray<FName> SavedNames;

	for(auto Name : AnchorNames)
	{
		SavedNames.Add(Name.name);
	}
	
	return SavedNames;
}

FName USpacesRuntimeBlueprintLibrary::GetSavedAnchorNameForARPin(UARPin* Pin)
{
	if(!IsValid(Pin) || Pin->GetNativeResource() == nullptr) return "None";
	
	const FSpacesSpatialAnchor* Anchor = static_cast<FSpacesSpatialAnchor*>(Pin->GetNativeResource());

	if(Anchor == nullptr) return "None";

	return Anchor->GetSavedAnchorName();
}

//Cached Component Versions to store data from xrGetComponentVersionsQCOM
static TOptional<TArray<FSpacesComponentVersion>> CachedComponentVersions;

bool USpacesRuntimeBlueprintLibrary::GetSpacesComponentVersions(TArray<FSpacesComponentVersion>& OutComponentVersions)
{
	if(CachedComponentVersions.IsSet())
	{
		OutComponentVersions = CachedComponentVersions.GetValue();
		FString logCV = "Enumerating component information from Spaces Services:\n";
		for(FSpacesComponentVersion ComponentVersion : OutComponentVersions)
		{
			logCV.Append("\nComponent '").Append(ComponentVersion.Name).Append("'\n     Version Identifier: ").Append(ComponentVersion.VersionId).Append("\n     Build Identifier: ").Append(ComponentVersion.BuildId).Append("\n     Build Date Time: ").Append(ComponentVersion.BuildDateTime);
		}
		UE_LOG(LogSpacesRuntime, Log, TEXT("%s"), *logCV);
		return true;
	}
	
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::Core))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces Core extension disabled."));
		return false;
	}
	
	const FSpacesCoreExtension* SpacesBase = static_cast<const FSpacesCoreExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::Core].Get());
	if(SpacesBase == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces Core extension disabled."));
		return false;
	}

	const bool bValidComponentVersions =  SpacesBase->GetComponentVersions(OutComponentVersions);
	if(bValidComponentVersions)
	{
		CachedComponentVersions = OutComponentVersions;
	}
	return bValidComponentVersions;
}

ESpacesControllerType USpacesRuntimeBlueprintLibrary::GetSpacesController()
{
#if PLATFORM_ANDROID	
	FSpacesSettingsWrapper SpacesSettings;
	return static_cast<ESpacesControllerType>(SpacesSettings.GetSpacesController());
#else  //PLATFORM_ANDROID
	if(const USpacesGameUserSettings* userSettings = Cast<USpacesGameUserSettings>(GEngine->GetGameUserSettings()))
	{
		return static_cast<ESpacesControllerType>(userSettings->GetSpacesController());
	}
	else
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Failed to cast to USpacesGameUserSettings"));
	}
#endif
	return ESpacesControllerType::ECT_GazeController;
}

void USpacesRuntimeBlueprintLibrary::SetSpacesController(ESpacesControllerType controller)
{
#if PLATFORM_ANDROID	
	FSpacesSettingsWrapper SpacesSettings;
	SpacesSettings.SetSpacesController(static_cast<int>(controller));
#else  //PLATFORM_ANDROID
	if(USpacesGameUserSettings* userSettings = Cast<USpacesGameUserSettings>(GEngine->GetGameUserSettings()))
	{
		userSettings->SetSpacesController(static_cast<int>(controller));
		userSettings->SaveSettings();
	}
	else
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Failed to cast to USpacesGameUserSettings"));
	}
#endif
}

void USpacesRuntimeBlueprintLibrary::ResetPose()
{
#if PLATFORM_ANDROID
	if(const USpacesRuntimeSettings* const SpacesSettings = GetDefault<const USpacesRuntimeSettings>())
	{
		if(SpacesSettings->bLaunchControllerOnHost)
		{
			FSpacesControllerWrapper SpacesPose;
			SpacesPose.ResetPose();
		}
	}
#endif  //PLATFORM_ANDROID
}

FString USpacesRuntimeBlueprintLibrary::GetEngineBuildDate()
{
	return FApp::GetBuildDate();
}

void USpacesRuntimeBlueprintLibrary::RenderSpatialMesh(AActor*& MeshActor, UMaterialInterface* Material, int& PreviousNumberOfIndices)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::Meshing))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces Meshing extension disabled."));
		return;
	}

	FSpacesMeshingExtension* SpacesMeshingExtension = static_cast<FSpacesMeshingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::Meshing].Get());
	UProceduralMeshComponent* ProceduralMesh = nullptr;
	TArray<FTrackedMeshHolder> Meshes = SpacesMeshingExtension->GetMeshes();
	for(FTrackedMeshHolder Mesh : Meshes)
	{
		if(MeshActor == nullptr)
		{
			MeshActor = NewObject<AActor>(GWorld->GetCurrentLevel(), FName("MeshingActor"));
			ProceduralMesh = NewObject<UProceduralMeshComponent>(MeshActor, UProceduralMeshComponent::StaticClass());
			ProceduralMesh->RegisterComponentWithWorld(GWorld);
			ProceduralMesh->SetCollisionProfileName(UCollisionProfile::NoCollision_ProfileName);
			ProceduralMesh->CreateMeshSection_LinearColor(0, Mesh.Vertices, (TArray<int32>)Mesh.Indices, Mesh.Normals, {}, {}, {}, false);
			ProceduralMesh->SetMaterial(0, Material);
			PreviousNumberOfIndices = Mesh.Indices.Num();
			UE_LOG(LogSpacesRuntime, Warning, TEXT("Creating MeshActor."));
		}else
		{
			ProceduralMesh = Cast<UProceduralMeshComponent>(MeshActor->GetComponentByClass(UProceduralMeshComponent::StaticClass()));
			if (ProceduralMesh)
			{
				if(Mesh.Indices.Num() != PreviousNumberOfIndices)
				{
					ProceduralMesh->CreateMeshSection_LinearColor(0, Mesh.Vertices, (TArray<int32>)Mesh.Indices, Mesh.Normals, {}, {}, {}, false);
					PreviousNumberOfIndices = Mesh.Indices.Num();
					ProceduralMesh->SetMaterial(0, Material);
					UE_LOG(LogSpacesRuntime, Warning, TEXT("Recreating ProceduralMesh section 0."));
				}else
				{
					ProceduralMesh->UpdateMeshSection_LinearColor(0, Mesh.Vertices, Mesh.Normals, {}, {}, {});
					UE_LOG(LogSpacesRuntime, Warning, TEXT("Getting and updating ProceduralmeshComponent."));
				}
			}
		}
		if (ProceduralMesh)
		{
			if(GEngine == nullptr) return;
			if(GEngine->XRSystem == nullptr) return;
			ProceduralMesh->SetWorldTransform(GEngine->XRSystem->GetTrackingToWorldTransform());
		}
	}
}

void USpacesRuntimeBlueprintLibrary::SetSpatialMeshingNormalsCalculationState(bool Active)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::Meshing))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces Meshing extension disabled."));
		return;
	}

	FSpacesMeshingExtension* SpacesMeshingExtension = static_cast<FSpacesMeshingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::Meshing].Get());

	SpacesMeshingExtension->SetComputeNormalsState(Active);
}

bool USpacesRuntimeBlueprintLibrary::GetSpatialMeshingNormalsCalculationState()
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::Meshing))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces Meshing extension disabled."));
		return false;
	}

	const FSpacesMeshingExtension* SpacesMeshingExtension = static_cast<FSpacesMeshingExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::Meshing].Get());

	return SpacesMeshingExtension->GetComputeNormalsState();
}

bool USpacesRuntimeBlueprintLibrary::IsExtensionLoaded(ESpacesExtensionBP Extension)
{
	const ESpacesExtension BaseExtension = static_cast<ESpacesExtension>(Extension); //Be catiuous with this, must have same order ESpacesExtension and ESpacesExtensionBP
	return FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(BaseExtension);
}

void USpacesRuntimeBlueprintLibrary::SetSpacesHapticsByValue(APlayerController* PlayerController, float Frequency, float Amplitude, float Duration, EControllerHand Hand)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::InputHostController))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces InputHostController extension disabled."));
		return;
	}
	
	FSpacesInputHostControllerExtension* InputHostControllerExtension = static_cast<FSpacesInputHostControllerExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::InputHostController].Get());

	if(InputHostControllerExtension != nullptr)
	{
		InputHostControllerExtension->SetHapticsDuration(Duration);
		InputHostControllerExtension->SetHapticsFrequency(Frequency);
		InputHostControllerExtension->SetHapticsAmplitude(Amplitude);
	}

	PlayerController->SetHapticsByValue(Frequency, Amplitude, Hand);
}

void USpacesRuntimeBlueprintLibrary::SetCameraFrameAccessState(bool Active)
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::CameraAccess))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces CameraAccess extension disabled."));
		return;
	}

	FSpacesCameraAccessExtension* CameraAccessExtension = static_cast<FSpacesCameraAccessExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::CameraAccess].Get());

	CameraAccessExtension->SetCameraAccessState(Active);
}

bool USpacesRuntimeBlueprintLibrary::IsCameraFrameAccessSupported()
{
	if(!FSpacesRuntimeModule::Get().GetLoadedExtensions().Contains(ESpacesExtension::CameraAccess))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Spaces CameraAccess extension disabled."));
		return false;
	}

	const FSpacesCameraAccessExtension* CameraAccessExtension = static_cast<FSpacesCameraAccessExtension*>(FSpacesRuntimeModule::Get().GetLoadedExtensions()[ESpacesExtension::CameraAccess].Get());

	return CameraAccessExtension->GetCameraFrameConfigFound();
}