﻿/******************************************************************************
 * File: SpacesXrFunction.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#pragma once

#include "CoreMinimal.h"
#include "External/openxr/openxr.h"
#include "SpacesCommon.h"

#include "OpenXRCore.h"

#define SPACES_OPENXR_FUNCTION_TYPES(_)				\
	/*SpacesCoreExtension*/							\
	_(xrConvertTimespecTimeToTimeKHR)				\
	_(xrConvertTimeToTimespecTimeKHR)				\
	_(xrGetComponentVersionsQCOM)					\
													\
	/*SpacesSpatialAnchorExtension*/				\
	_(xrCreateSpatialAnchorMSFT)					\
	_(xrCreateSpatialAnchorSpaceMSFT)				\
	_(xrDestroySpatialAnchorMSFT)					\
	_(xrDestroySpace)								\
	_(xrLocateSpace)								\
													\
	_(xrCreateSpatialAnchorStoreConnectionMSFT)		\
	_(xrDestroySpatialAnchorStoreConnectionMSFT)	\
	_(xrPersistSpatialAnchorMSFT)					\
	_(xrUnpersistSpatialAnchorMSFT)					\
	_(xrEnumeratePersistedSpatialAnchorNamesMSFT)	\
	_(xrCreateSpatialAnchorFromPersistedNameMSFT)	\
	_(xrClearSpatialAnchorStoreMSFT)				\
													\
	/*SpacesPlaneDetectionExtension*/				\
	_(xrCreatePlaneDetectionQCOM)					\
	_(xrDestroyPlaneDetectionQCOM)					\
	_(xrLocatePlanesQCOM)							\
	_(xrGetPlaneDetectionStateQCOM)					\
	_(xrGetPlaneConvexHullVertexBufferQCOM)			\
													\
	/*SpacesImageTrackingExtension*/				\
	_(xrCreateImageTrackerQCOM)						\
	_(xrDestroyImageTrackerQCOM)					\
	_(xrLocateImageTargetsQCOM)						\
	_(xrImageTargetToNameAndIdQCOM)					\
	_(xrSetTrackingOptimizationSettingsHintQCOM)	\
													\
	/*SpacesHitTestingExtension*/					\
	_(xrCreateRayCastQCOM)							\
	_(xrEnumerateRayCastTargetTypesQCOM)			\
	_(xrGetRayCastStateQCOM)						\
	_(xrGetRayCollisionsQCOM)						\
	_(xrCastRayQCOM)								\
	_(xrDestroyRayCastQCOM)							\
													\
	/*SpacesHandTrackingExtension*/					\
	_(xrCreateHandTrackerEXT)						\
	_(xrDestroyHandTrackerEXT)						\
	_(xrLocateHandJointsEXT)						\
	_(xrGetHandGestureQCOM)							\
	_(xrCreateHandMeshSpaceMSFT)					\
	_(xrUpdateHandMeshMSFT)							\
													\
	/*SpacesMeshingExtension*/						\
	_(xrCreateSceneObserverMSFT)					\
	_(xrDestroySceneObserverMSFT)					\
	_(xrCreateSceneMSFT)							\
	_(xrDestroySceneMSFT)							\
	_(xrEnumerateSceneComputeFeaturesMSFT)			\
	_(xrComputeNewSceneMSFT)						\
	_(xrGetSceneComponentsMSFT)						\
	_(xrLocateSceneComponentsMSFT)					\
	_(xrGetSceneMeshBuffersMSFT)					\
	_(xrGetSceneComputeStateMSFT)					\
													\
	/*SpacesCameraAccessExtension*/					\
	_(xrEnumerateCamerasQCOMX)						\
	_(xrCreateCameraHandleQCOMX)					\
	_(xrGetSupportedFrameConfigurationsQCOMX)		\
	_(xrReleaseCameraHandleQCOMX)					\
	_(xrAccessFrameQCOMX)							\
	_(xrReleaseFrameQCOMX)							\
	
	
/***
 * @brief Defines an entrypoint pointer type
 * @param Name OpenXR entrypoint name
 * @param ... Entrypoint parameters type
 */
#define DEFINE_XR_ENTRYPOINT_PTR(Name, ...) typedef XrResult (XRAPI_PTR* PFN_##Name)(__VA_ARGS__);

/***
 * @brief Declares an external entrypoint pointer variable that should be defined in the module
 * @details Must be used inside SnapdragonSpaces namespace
 * @details Must be defined somewhere in the module using DEFINE_XR_ENTRYPOINT
 * @param Name OpenXR entrypoint name
 */
#define DECLARE_XR_ENTRYPOINT(Name)									\
	using F##Name = TXrFunction<SnapdragonSpaces::F##Name##Info>;	\
	extern F##Name Name;

/***
 * @brief Defines an entrypoint pointer variable and initialize it with XR_NULL_HANDLE
 * @details Must get the entrypoint handle in order to be used, the pointer type of the method needs to be defined,
 * either by an external header or by the usage of DEFINE_XR_ENTRYPOINT_PTR
 * @param Name OpenXR entrypoint name
 */
#define DEFINE_XR_ENTRYPOINT(Name)												\
SnapdragonSpaces::F##Name SnapdragonSpaces::Name = SnapdragonSpaces::F##Name();

/***
 * @brief Tries to get an entrypoint handle and stores it in it's pointer variable
 * @details The entrypoint needs to be defined using DEFINE_XR_ENTRYPOINT
 * @param Name OpenXR entrypoint name
 */
#define GET_XR_ENTRYPOINT(Name)	SnapdragonSpaces::Name = SnapdragonSpaces::F##Name(GetInstanceHandle());


#define DEFINE_SPACES_OPENXR_FUNCTION_TYPE(NativeType)		\
	struct F##NativeType##Info{								\
		using Type = PFN_##NativeType;						\
		static constexpr TCHAR Name[] = TEXT(#NativeType);	\
	};														\
	constexpr TCHAR F##NativeType##Info::Name[];


namespace SnapdragonSpaces
{

	/***
	 * @brief Helper class to manage OpenXR functions
	 * @details Used to wrap an OpenXR function and handle the calls to it
	 * @param PFN Function pointer type
	 */
	template<typename SpacesFunctionType>
	class TXrFunction
	{

	public:

		using FunctionType = SpacesFunctionType;
		using XRType = typename FunctionType::Type;

		TXrFunction(){};
		
		TXrFunction(XrInstance Instance)
		{
			if(XR_FAILED(OpenXRDynamicAPI::xrGetInstanceProcAddr(						
				Instance,														
				TCHAR_TO_UTF8(SpacesFunctionType::Name),																		
				reinterpret_cast<PFN_xrVoidFunction*>(&XrFunction)
			)))
			{
				UE_LOG(LogSpacesRuntime, Error, TEXT("Failed to initialize %s"), SpacesFunctionType::Name);
				XrFunction = nullptr;
			}
			UE_LOG(LogSpacesRuntime, Log, TEXT("Initialized %s (%p)"), SpacesFunctionType::Name, XrFunction);
		}
		
		template<typename... TArgs>
		bool operator()(TArgs... Args) const
		{
			if(!IsValid())
			{
				UE_LOG(LogSpacesRuntime, Warning, TEXT("Trying to use uninitialized %s"), SpacesFunctionType::Name);
				return false;
			}
			LastResult = XrFunction(Forward<TArgs>(Args)...);
			return XR_SUCCEEDED(LastResult);
		}

		XrResult GetLastResult() const noexcept
		{
			return LastResult;
		}

		bool IsValid() const noexcept
		{
			return XrFunction != nullptr;
		}

	private:

		XRType XrFunction = nullptr;
		mutable XrResult LastResult = XR_ERROR_VALIDATION_FAILURE;
		
	};
}

namespace SnapdragonSpaces
{
	SPACES_OPENXR_FUNCTION_TYPES(DEFINE_SPACES_OPENXR_FUNCTION_TYPE)
	SPACES_OPENXR_FUNCTION_TYPES(DECLARE_XR_ENTRYPOINT)
};
