﻿/******************************************************************************
 * File: SpacesXrHandle.h
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#pragma once

#include "CoreMinimal.h"
#include "External/openxr/openxr.h"
#include "SpacesCommon.h"

#define SPACES_OPENXR_HANDLE_TYPES(_)																											\
	/*SpacesSpatialAnchorExtension*/																											\
	_(SpatialAnchorMSFT, XrSpatialAnchorMSFT, &SnapdragonSpaces::xrCreateSpatialAnchorMSFT, &SnapdragonSpaces::xrDestroySpatialAnchorMSFT)		\
	_(SpatialAnchorSpaceMSFT, XrSpace, &SnapdragonSpaces::xrCreateSpatialAnchorSpaceMSFT, &SnapdragonSpaces::xrDestroySpace)					\
																																				\
	/*SpacesPlaneDetectionExtension*/																											\
	_(PlaneDetectionQCOM, XrPlaneDetectionQCOM, &SnapdragonSpaces::xrCreatePlaneDetectionQCOM, &SnapdragonSpaces::xrDestroyPlaneDetectionQCOM)	\
																																				\
	/*SpacesImageTrackingExtension*/																											\
	_(ImageTrackerQCOM, XrImageTrackerQCOM, &SnapdragonSpaces::xrCreateImageTrackerQCOM, &SnapdragonSpaces::xrDestroyImageTrackerQCOM)			\
																																				\
	/*SpacesHitTestingExtension*/																												\
	_(RayCastQCOM, XrRayCastQCOM, &SnapdragonSpaces::xrCreateRayCastQCOM, &SnapdragonSpaces::xrDestroyRayCastQCOM)								\
																																				\
	/*SpacesMeshingExtension*/																													\
	_(SceneObserverMSFT, XrSceneObserverMSFT, &SnapdragonSpaces::xrCreateSceneObserverMSFT, &SnapdragonSpaces::xrDestroySceneObserverMSFT)		\
	_(SceneMSFT, XrSceneMSFT, &SnapdragonSpaces::xrCreateSceneMSFT, &SnapdragonSpaces::xrDestroySceneMSFT)										\



/***
 * @brief Defines a handle type
 * @details Macro to quickly define a handle wrapper using only its name.
 * Designed for handles with matching "xrCreate" and "xrDestroy" functions
 * @details Uses Xr<Name> as handle type and xrCreate<Name> and xrDestroy<Name> for
 * managing the handle resource
 * @param Name Handle type name
 */
#define DEFINE_XR_HANDLE(Name, NativeType, ...)						\
	using F##Name##Handle = TXrHandle<F##Name##Info, __VA_ARGS__>;


#define DEFINE_SPACES_OPENXR_HANDLE_TYPE(DisplayName, NativeType, ...)	\
	struct F##DisplayName##Info{										\
		using Type = NativeType;										\
		static constexpr TCHAR Name[] = TEXT(#DisplayName);				\
	};																	\
	constexpr TCHAR F##DisplayName##Info::Name[];


namespace SnapdragonSpaces
{

	/***
	 * @brief Helper class to manage OpenXR handles
	 * @details Used to wrap handles that uses xrCreate and xrDestroy functions
	 * @param XRType Handle type
	 * @param xrCreateFunction Reference to the static variable with a pointer to the xrCreate function
	 * @param xrDestroyFunction Reference to the static variable with a pointer to the xrDestroy function
	 */
	template<typename SpacesHandleType, auto* xrCreateFunction, auto* xrDestroyFunction>
	class TXrHandle
	{
			
	public:

		using HandleType = SpacesHandleType;
		using XRType = typename HandleType::Type;

		template<typename... TArgs>
		TXrHandle(TArgs... Args) noexcept
		{
			
			if(!xrCreateFunction->IsValid())
			{
				UE_LOG(LogSpacesRuntime, Warning, TEXT("[%s] xrCreateFunction is not valid"), HandleType::Name);
				return;
			}
			if(!xrDestroyFunction->IsValid())
			{
				UE_LOG(LogSpacesRuntime, Warning, TEXT("[%s] xrDestroyFunction is not valid"), HandleType::Name);
				return;
			}
			
			if((*xrCreateFunction)(Forward<TArgs>(Args)..., &Handle) == false)
			{
				UE_LOG(LogSpacesRuntime, Warning, TEXT("[%s] xrCreateFunction failed: %s"), HandleType::Name, SnapdragonSpaces::XrEnumStr(xrCreateFunction->GetLastResult()));
				Handle = XR_NULL_HANDLE;
				return;
			}
			bInitialized = true;
		}
		
		TXrHandle(XRType&& InitializedHandle) noexcept
		{
			Handle = InitializedHandle;
			InitializedHandle = XR_NULL_HANDLE;
			bInitialized = true;
		}
		
		TXrHandle(TXrHandle&& Other) noexcept
		{
			*this = MoveTemp(Other);
		}
		
		TXrHandle& operator=(TXrHandle&& Other) noexcept
		{
			if(Handle != Other.Handle)
			{
				if(IsValid())
				{
					Clear();
				}
				Handle = Other.Handle;
				bInitialized = Other.bInitialized;
			}
			Other.Invalidate();
			return *this;
		}
		
		TXrHandle() = default;
		
		TXrHandle(const TXrHandle&) = delete;
		TXrHandle& operator=(const TXrHandle&) = delete;
		
		~TXrHandle() noexcept
		{
			if(IsValid())
			{
				Clear();
			}
		}
		
		void Clear()
		{
			if(!IsValid())
			{
				UE_LOG(LogSpacesRuntime, Warning, TEXT("[%s] Handle is not valid"), HandleType::Name);
				return;
			}
			
			ensureMsgf(xrDestroyFunction->IsValid(), TEXT("[%s] xrDestroyFunction is not valid"), HandleType::Name);
			
			(*xrDestroyFunction)(Handle);

			checkf(xrDestroyFunction->GetLastResult() != XR_ERROR_RUNTIME_FAILURE, TEXT("[%s] Clear function failed, runtime failure!"), HandleType::Name);
			Invalidate();
		}

		bool IsValid() const noexcept
		{
			return bInitialized;
		}
		
		operator bool() const noexcept
		{
			return IsValid();
		}

		XRType Value() const noexcept
		{
			return Handle;
		}
		
		XRType operator *() const noexcept
		{
			return Value();
		}

	private:

		void Invalidate() noexcept
		{
			Handle = XR_NULL_HANDLE;
			bInitialized = false;
		}
		
		XRType Handle = XR_NULL_HANDLE;
		bool bInitialized = false;
		
	};
	
	template<typename SpacesHandleType>
	class TXrHandle<SpacesHandleType, (void*)nullptr, (void*)nullptr>
	{
	public:

		using HandleType = SpacesHandleType;
		using XRType = typename HandleType::Type;
		
		TXrHandle() = default;
		TXrHandle(XRType InitializedHandle)
		{
			Handle = InitializedHandle;
		}

		bool IsValid() const noexcept
		{
			return Handle != XR_NULL_HANDLE;
		}
		
		operator bool() const noexcept
		{
			return IsValid();
		}

		XRType Value() const noexcept
		{
			return Handle;
		}
		
		XRType operator *() const noexcept
		{
			return Value();
		}
		
	private:
		
		XRType Handle = XR_NULL_HANDLE;
		
	};
	
};

namespace SnapdragonSpaces
{
	SPACES_OPENXR_HANDLE_TYPES(DEFINE_SPACES_OPENXR_HANDLE_TYPE)
	SPACES_OPENXR_HANDLE_TYPES(DEFINE_XR_HANDLE)
}
