﻿/******************************************************************************
 * File: SpacesSpatialAnchor.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesSpatialAnchor.h"

EARTrackingState FSpacesSpatialAnchor::GetUpdatedTransform(XrSpace TrackingSpace, XrTime DisplayTime, float worldToMeterScale, FTransform& Transform) const
{
	XrSpaceLocation SpaceLocation{
		XR_TYPE_SPACE_LOCATION
	};
	
	if(!SnapdragonSpaces::xrLocateSpace(AnchorSpace.Value(), TrackingSpace, DisplayTime, &SpaceLocation))
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("xrLocateSpace failed (%s)"), SnapdragonSpaces::XrEnumStr(SnapdragonSpaces::xrLocateSpace.GetLastResult()));
		return EARTrackingState::Unknown;
	}

	if(SnapdragonSpaces::ToARTrackingState(SpaceLocation.locationFlags) == EARTrackingState::StoppedTracking)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Not valid locationFlags"));
		return EARTrackingState::NotTracking;
	}
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("SpaceLocation.pose: [x:%f y:%f z:%f w:%f] [x:%f y:%f z:%f]"),
		   SpaceLocation.pose.orientation.x, SpaceLocation.pose.orientation.y,
		   SpaceLocation.pose.orientation.z, SpaceLocation.pose.orientation.w, SpaceLocation.pose.position.x,
		   SpaceLocation.pose.position.y, SpaceLocation.pose.position.z);
	Transform = ToFTransform(SpaceLocation.pose, worldToMeterScale);
	UE_LOG(LogSpacesRuntime, Log, TEXT("xrLocateSpace to Transform: %s"), *(Transform.ToString()));
		
	return EARTrackingState::Tracking;//SnapdragonSpaces::ToARTrackingState(SpaceLocation.locationFlags);
}

XrSpatialAnchorMSFT FSpacesSpatialAnchor::GetAnchor() const
{
    return Anchor.Value();
}

XrSpace FSpacesSpatialAnchor::GetAnchorSpace() const
{
    return AnchorSpace.Value();
}

bool FSpacesSpatialAnchor::IsValid() const
{
	return Anchor.IsValid() && AnchorSpace.IsValid();
}

FSpacesSpatialAnchor::FSpacesSpatialAnchor(XrInstance InInstance, XrSession InSession, XrSpace TrackingSpace, XrTime DisplayTime,
                                           XrPosef InPose)
{	
	UE_LOG(LogSpacesRuntime, Log, TEXT("Instance: %llu"), InInstance);
	UE_LOG(LogSpacesRuntime, Log, TEXT("InSession: %llu"), InSession);
	UE_LOG(LogSpacesRuntime, Log, TEXT("DisplayTime: %lld"), DisplayTime);

	
    const XrSpatialAnchorCreateInfoMSFT AnchorCreateInfo{
        XR_TYPE_SPATIAL_ANCHOR_CREATE_INFO_MSFT,
        nullptr,
        TrackingSpace,
        InPose,
        DisplayTime
    };

	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorCreateInfo.type: %s"), SnapdragonSpaces::XrEnumStr(AnchorCreateInfo.type));
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorCreateInfo.next: %p"), AnchorCreateInfo.next);
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorCreateInfo.space: %llu"), AnchorCreateInfo.space);
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorCreateInfo.pose: [x:%f y:%f z:%f w:%f] [x:%f y:%f z:%f]"),
	       AnchorCreateInfo.pose.orientation.x, AnchorCreateInfo.pose.orientation.y,
	       AnchorCreateInfo.pose.orientation.z, AnchorCreateInfo.pose.orientation.w, AnchorCreateInfo.pose.position.x,
	       AnchorCreateInfo.pose.position.y, AnchorCreateInfo.pose.position.z);
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorCreateInfo.time: %lld"), AnchorCreateInfo.time);


	Anchor = SnapdragonSpaces::FSpatialAnchorMSFTHandle(InSession, &AnchorCreateInfo);
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("Returned anchor handle: %llu"), Anchor.Value());

    if(!Anchor.IsValid())
    {
		return;
    }
    
    const XrSpatialAnchorSpaceCreateInfoMSFT AnchorSpaceCreateInfo{
        XR_TYPE_SPATIAL_ANCHOR_SPACE_CREATE_INFO_MSFT,
        nullptr,
        Anchor.Value(),
		XrPosef{ XrQuaternionf { 0.0f, 0.0f, 0.0f, 1.0f}, XrVector3f { 0.0f, 0.0f, 0.0f } }
    };
    
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorSpaceCreateInfo.type: %s"), SnapdragonSpaces::XrEnumStr(AnchorSpaceCreateInfo.type));
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorSpaceCreateInfo.next: %p"), AnchorSpaceCreateInfo.next);
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorSpaceCreateInfo.anchor: %llu"), AnchorSpaceCreateInfo.anchor);
	UE_LOG(LogSpacesRuntime, Log, TEXT("AnchorSpaceCreateInfo.poseInAnchorSpace: [x:%f y:%f z:%f w:%f] [x:%f y:%f z:%f]"),
	       AnchorSpaceCreateInfo.poseInAnchorSpace.orientation.x, AnchorSpaceCreateInfo.poseInAnchorSpace.orientation.y,
	       AnchorSpaceCreateInfo.poseInAnchorSpace.orientation.z, AnchorSpaceCreateInfo.poseInAnchorSpace.orientation.w,
	       AnchorSpaceCreateInfo.poseInAnchorSpace.position.x, AnchorSpaceCreateInfo.poseInAnchorSpace.position.y,
	       AnchorSpaceCreateInfo.poseInAnchorSpace.position.z);
	
	AnchorSpace = SnapdragonSpaces::FSpatialAnchorSpaceMSFTHandle(InSession, &AnchorSpaceCreateInfo);
	UE_LOG(LogSpacesRuntime, Log, TEXT("Returned anchor space handle: %llu"), AnchorSpace.Value());
    
	if(!AnchorSpace.IsValid())
	{
		Anchor.Clear();
		return;
	}
}

FSpacesSpatialAnchor::FSpacesSpatialAnchor(XrInstance InInstance, XrSession InSession, XrSpatialAnchorMSFT SpatialAnchor)
{
	Anchor= MoveTemp(SpatialAnchor);
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("Returned anchor handle: %llu"), Anchor.Value());

	if(!Anchor.IsValid())
	{
		return;
	}
    
	const XrSpatialAnchorSpaceCreateInfoMSFT AnchorSpaceCreateInfo{
		XR_TYPE_SPATIAL_ANCHOR_SPACE_CREATE_INFO_MSFT,
		nullptr,
		Anchor.Value(),
		XrPosef{ XrQuaternionf { 0.0f, 0.0f, 0.0f, 1.0f}, XrVector3f { 0.0f, 0.0f, 0.0f } }
	};

	AnchorSpace = SnapdragonSpaces::FSpatialAnchorSpaceMSFTHandle(InSession, &AnchorSpaceCreateInfo);
	UE_LOG(LogSpacesRuntime, Log, TEXT("Returned anchor space handle: %llu"), AnchorSpace.Value());
    
	if(!AnchorSpace.IsValid())
	{
		Anchor.Clear();
		return;
	}
}

FSpacesSpatialAnchor::~FSpacesSpatialAnchor()
{
	AnchorSpace.Clear();
	Anchor.Clear();
}

FSpacesSpatialAnchor::FSpacesSpatialAnchor(FSpacesSpatialAnchor&& Other) noexcept
{
	*this = MoveTemp(Other);
}

FSpacesSpatialAnchor& FSpacesSpatialAnchor::operator=(FSpacesSpatialAnchor&& Other) noexcept
{
	AnchorSpace = MoveTemp(Other.AnchorSpace);
	Anchor = MoveTemp(Other.Anchor);
	return *this;
}

bool FSpacesSpatialAnchor::Save(XrSpatialAnchorStoreConnectionMSFT AnchorStore, XrSpatialAnchorPersistenceNameMSFT PersistenceName) const
{
	if(!Anchor.IsValid())
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Anchor handle not found"));
		return false;
	}

	XrSpatialAnchorPersistenceInfoMSFT Info {
		XR_TYPE_SPATIAL_ANCHOR_PERSISTENCE_INFO_MSFT,
		nullptr,
		PersistenceName,
		Anchor.Value()
	};

	
	if(SnapdragonSpaces::xrPersistSpatialAnchorMSFT(AnchorStore, &Info) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed adding the spatial anchor %s in the spatial anchor store"), *FString(PersistenceName.name));
		return false;
	}

	UE_LOG(LogSpacesRuntime, Log, TEXT("Persist a spatial anchor %s in the spatial anchor store"), *FString(PersistenceName.name));
	return true;
}

FName FSpacesSpatialAnchor::GetSavedAnchorName() const
{
	return SavedAnchorName;
}
