/******************************************************************************
 * File: SpacesSpatialAnchorExtension.cpp
 * Copyright (c) 2022 Qualcomm Technologies, Inc. and/or its subsidiaries. All rights reserved.
 *
 * Confidential and Proprietary - Qualcomm Technologies, Inc.
 *
 */

#include "SpacesSpatialAnchorExtension.h"
#include "SpacesSpatialAnchor.h"
#include "ARPin.h"

#include "Runtime/Launch/Resources/Version.h"
#if ENGINE_MAJOR_VERSION == 5
#include "IOpenXRHMDModule.h"
#else
#include "IOpenXRHMDPlugin.h"
#endif

#include "OpenXRCore.h"
#include "Engine/Engine.h"
#include "SubSystems/SpacesEngineSubsystem.h"
#include "Async/Async.h"

#define LOCTEXT_NAMESPACE "FSpacesSpatialAnchorExtension"

bool FSpacesSpatialAnchorExtension::OnPinComponent(UARPin* Pin, XrSession InSession, XrSpace TrackingSpace, XrTime DisplayTime, float worldToMeterScale)
{

	UE_LOG(LogSpacesRuntime, Log, TEXT("OnPinComponent"));
	

	UE_LOG(LogSpacesRuntime, Log, TEXT("Pin: %s"), *(Pin->GetLocalToTrackingTransform().ToString()));
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("LocalToTrackingTransform: %s"), *(Pin->GetLocalToTrackingTransform().ToString()));
	UE_LOG(LogSpacesRuntime, Log, TEXT("LocalToTrackingTransform_NoAlignment: %s"), *(Pin->GetLocalToTrackingTransform_NoAlignment().ToString()));
	

	FSpacesSpatialAnchor* Anchor = new FSpacesSpatialAnchor(GetInstanceHandle(), InSession, TrackingSpace, DisplayTime, ToXrPose(Pin->GetLocalToTrackingTransform(), worldToMeterScale));

    if(!Anchor->IsValid())
    {
		Pin->OnTrackingStateChanged(EARTrackingState::Unknown);
    	delete Anchor;
		UE_LOG(LogSpacesRuntime, Log, TEXT("OnPinComponent failed, anchor not valid"));
        return false;
    }
	
    Pin->SetNativeResource(Anchor);
    return true;
}

void FSpacesSpatialAnchorExtension::OnRemovePin(UARPin* Pin)
{
	
	const FSpacesSpatialAnchor* Anchor = static_cast<FSpacesSpatialAnchor*>(Pin->GetNativeResource());
	
	if(Anchor == nullptr){
		Pin->OnTrackingStateChanged(EARTrackingState::Unknown);
		return;
	}
	
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnRemovePin (%llu)"), Anchor->GetAnchor());
	
    delete Anchor;
    Pin->OnTrackingStateChanged(EARTrackingState::StoppedTracking);
    Pin->SetNativeResource(nullptr);
}

void FSpacesSpatialAnchorExtension::OnUpdatePin(UARPin* Pin, XrSession InSession, XrSpace TrackingSpace, XrTime DisplayTime, float worldToMeterScale)
{
	const FSpacesSpatialAnchor* Anchor = static_cast<FSpacesSpatialAnchor*>(Pin->GetNativeResource());
	
    if(Anchor == nullptr)
    {
        Pin->OnTrackingStateChanged(EARTrackingState::Unknown);
    	return;
    }
    	
	UE_LOG(LogSpacesRuntime, Log, TEXT("OnUpdatePin (%llu)"), Anchor->GetAnchor());

	FTransform UpdatedTransform;
	const EARTrackingState UpdatedTrackingState = Anchor->GetUpdatedTransform(TrackingSpace, DisplayTime, worldToMeterScale, UpdatedTransform);
	
	if(!UpdatedTransform.Equals(Pin->GetLocalToTrackingTransform_NoAlignment()) &&
		UpdatedTrackingState == EARTrackingState::Tracking)
	{
		Pin->OnTransformUpdated(UpdatedTransform);
	}
	
	if( UpdatedTrackingState != Pin->GetTrackingState() &&
		(UpdatedTrackingState != EARTrackingState::StoppedTracking || Pin->GetTrackingState() != EARTrackingState::Unknown))
	{
		Pin->OnTrackingStateChanged(UpdatedTrackingState);
	}
}

bool FSpacesSpatialAnchorExtension::IsLocalPinSaveSupported() const
{
#if ENGINE_MAJOR_VERSION == 5
	return IOpenXRHMDModule::Get().IsExtensionEnabled(XR_MSFT_SPATIAL_ANCHOR_PERSISTENCE_EXTENSION_NAME);
#else
	return IOpenXRHMDPlugin::Get().IsExtensionEnabled(XR_MSFT_SPATIAL_ANCHOR_PERSISTENCE_EXTENSION_NAME);
#endif
}

bool FSpacesSpatialAnchorExtension::ArePinsReadyToLoad()
{
    return AnchorStore != XR_NULL_HANDLE;
}

void FSpacesSpatialAnchorExtension::LoadARPins(XrSession InSession, TFunction<UARPin*(FName)> OnCreatePin)
{
	if(AnchorLoadingFuture.IsValid()) return;
	
	TArray<XrSpatialAnchorPersistenceNameMSFT> AnchorNames = GetSavedAnchorNames();
	if(AnchorNames.Num() == 0) return;

	TMap<FName, TWeakObjectPtr<UARPin>> PinToLoad;
	UE_LOG(LogSpacesRuntime, Log, TEXT("Preparing %i ARPins to load"), AnchorNames.Num());
	for(const XrSpatialAnchorPersistenceNameMSFT& AnchorName : AnchorNames)
	{
		FName name = FName(AnchorName.name);
		TWeakObjectPtr<UARPin> Pin = OnCreatePin(name);

		if(Pin != nullptr)
		{
			Pin->OnTrackingStateChanged(EARTrackingState::Unknown);
			PinToLoad.Add(FName(AnchorName.name), Pin);
		}
		else
		{
			UE_LOG(LogSpacesRuntime, Error, TEXT("Failed creating the pin %s."), *name.ToString());
		}

	}

	if(PinToLoad.Num() == 0)
	{
		UE_LOG(LogSpacesRuntime, Log, TEXT("Pin to load is empty."));
		return;
	}

	TPromise<TMap<TWeakObjectPtr<UARPin>, TSharedPtr<FSpacesSpatialAnchor, ESPMode::ThreadSafe>>> Promise;
	AnchorLoadingFuture = Promise.GetFuture();
	AsyncTask(
		ENamedThreads::AnyThread,
		[InInstance = GetInstanceHandle(), InSession = InSession, Promise = MoveTemp(Promise), AnchorStore = AnchorStore, PinToLoad = MoveTemp(PinToLoad)]() mutable 
		{
			Promise.SetValue(LoadAnchors(InInstance, InSession, AnchorStore, MoveTemp(PinToLoad)));
		}
	);
}

TMap<TWeakObjectPtr<UARPin>, TSharedPtr<FSpacesSpatialAnchor, ESPMode::ThreadSafe>> FSpacesSpatialAnchorExtension::LoadAnchors(XrInstance InInstance, XrSession InSession, XrSpatialAnchorStoreConnectionMSFT AnchorStore, TMap<FName, TWeakObjectPtr<UARPin>> PinToLoad)
{
	TMap<TWeakObjectPtr<UARPin>, TSharedPtr<FSpacesSpatialAnchor, ESPMode::ThreadSafe>> AnchorToLoad;

	for(auto& Item : PinToLoad)
	{
		XrSpatialAnchorPersistenceNameMSFT AnchorName;
		if(InitializePersistenceName(Item.Key, AnchorName))
		{
			XrSpatialAnchorMSFT SpatialAnchor;
			if(!Item.Value.IsExplicitlyNull() && CreateSpatialAnchorFromPersistenceName(InSession, AnchorStore, SpatialAnchor, AnchorName))
			{
				TSharedPtr<FSpacesSpatialAnchor, ESPMode::ThreadSafe> Anchor = MakeShared<FSpacesSpatialAnchor, ESPMode::ThreadSafe>(InInstance, InSession, SpatialAnchor);
				AnchorToLoad.Add(Item.Value, Anchor);
			}
			else
			{
				AnchorToLoad.Add(Item.Value, nullptr);
				UE_LOG(LogSpacesRuntime, Error, TEXT("Failed creating the spatial anchor %s"), *FName(AnchorName.name).ToString());
			}
		}
	}
	
	return AnchorToLoad;
}

bool FSpacesSpatialAnchorExtension::InitializePersistenceName(FName InName, XrSpatialAnchorPersistenceNameMSFT& PersistenceName)
{
	int32 NameLen = InName.ToString().Len() + 1;
	if(NameLen >= XR_MAX_SPATIAL_ANCHOR_NAME_SIZE_MSFT)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Anchor name is too long."));
		return false;
	}
	
	FPlatformString::Strncpy(PersistenceName.name, FTCHARToUTF8(*InName.ToString().ToLower()).Get(), NameLen);
	return true;
}

bool FSpacesSpatialAnchorExtension::CreateSpatialAnchorFromPersistenceName(XrSession InSession, XrSpatialAnchorStoreConnectionMSFT AnchorStore, XrSpatialAnchorMSFT& SpatialAnchor, XrSpatialAnchorPersistenceNameMSFT PersistenceName)
{
	XrSpatialAnchorFromPersistedAnchorCreateInfoMSFT CreateInfo {
		XR_TYPE_SPATIAL_ANCHOR_FROM_PERSISTED_ANCHOR_CREATE_INFO_MSFT,
		nullptr,
		AnchorStore,
		PersistenceName
	};

	if(SnapdragonSpaces::xrCreateSpatialAnchorFromPersistedNameMSFT(InSession, &CreateInfo, &SpatialAnchor) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed xrCreateSpatialAnchorFromPersistedNameMSFT (%s)"), SnapdragonSpaces::XrEnumStr(SnapdragonSpaces::xrCreateSpatialAnchorFromPersistedNameMSFT.GetLastResult()));
		return false;
	}

	return true;
}

bool FSpacesSpatialAnchorExtension::SaveARPin(XrSession InSession, FName InName, UARPin* InPin)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Trying to save an AR Pin (%s)"), *InName.ToString());
	if (!ArePinsReadyToLoad())
	{
		return false;
	}

	if(!IsValid(InPin))
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("AR Pin from %s is NULL"), *InName.ToString());
		return false;
	}
	
	AnchorPendingToSave.Add(InName, InPin);
	CreateAsyncSaveThread();

	return true;
}

bool FSpacesSpatialAnchorExtension::CreateAsyncSaveThread()
{
	if (AnchorSavingFuture.IsValid()) return false;
	if (AnchorPendingToSave.Num() == 0) return false;

	TArray<FName> KeyNames;
	AnchorPendingToSave.GetKeys(KeyNames);
	
	FName Name = KeyNames[0];
	UARPin* ARPin = *AnchorPendingToSave.Find(Name);

	AnchorSaving = Name;
	UE_LOG(LogSpacesRuntime, Log, TEXT("Saving anchor %s"), *AnchorSaving.ToString());
	TPromise<bool> Promise;
	AnchorSavingFuture = Promise.GetFuture();
	AsyncTask(
		ENamedThreads::AnyThread,
		[Promise = MoveTemp(Promise), AnchorStore = AnchorStore, InName = Name, InPin = MoveTemp(ARPin)]() mutable 
		{
			Promise.SetValue(SaveAnchor(AnchorStore, InName, MoveTemp(InPin)));
		}
	);

	return true;
}

bool FSpacesSpatialAnchorExtension::SaveAnchor(XrSpatialAnchorStoreConnectionMSFT AnchorStore, FName InName, UARPin* InPin)
{
	const FSpacesSpatialAnchor* Anchor = static_cast<FSpacesSpatialAnchor*>(InPin->GetNativeResource());
	
	if(Anchor == nullptr)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("AR Pin from %s not valid"), *InName.ToString());
		return false;
	}

	XrSpatialAnchorPersistenceNameMSFT PersistenceName;
	if(InitializePersistenceName(InName, PersistenceName))
	{
		return Anchor->Save(AnchorStore, PersistenceName);
	}

	return false;
}

void FSpacesSpatialAnchorExtension::RemoveSavedARPin(XrSession InSession, FName InName)
{
	if (!ArePinsReadyToLoad())
	{
		return;
	}
	
	XrSpatialAnchorPersistenceNameMSFT PersistenceName;
	if(!InitializePersistenceName(InName, PersistenceName)) return;
	
	if(SnapdragonSpaces::xrUnpersistSpatialAnchorMSFT(AnchorStore, &PersistenceName) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed removing the record of the anchor %s in the spatial anchor store"), *InName.ToString());
		return;
	} 

	UE_LOG(LogSpacesRuntime, Log, TEXT("Remove the record of the anchor %s in the spatial anchor store completed successfully"), *InName.ToString());
}

void FSpacesSpatialAnchorExtension::RemoveAllSavedARPins(XrSession InSession)
{
	if (!ArePinsReadyToLoad())
	{
		return;
	}
	
	//if(AnchorLoadingFuture.IsValid())
	//{
	//	AnchorLoadingFuture.Reset();
	//}

	if(AnchorSavingFuture.IsValid())
	{
		AnchorSavingFuture.Reset();
		AnchorPendingToSave.Empty();

		if(USpacesEngineSubsystem* SpacesSubsystem = GEngine->GetEngineSubsystem<USpacesEngineSubsystem>())
		{
			SpacesSubsystem->OnSpacesAnchorIsSaved.Broadcast(false, AnchorSaving);
		}
	}

	if(SnapdragonSpaces::xrClearSpatialAnchorStoreMSFT(AnchorStore) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed removing all spatial anchors from spatial anchor store"));
		return;
	}

	UE_LOG(LogSpacesRuntime, Log, TEXT("Removed all spatial anchors from spatial anchor store successfully"));
}

FSpacesSpatialAnchorExtension::FSpacesSpatialAnchorExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Startup SpacesSpatialAnchorExtension"));
	RegisterOpenXRExtensionModularFeature();

}

FSpacesSpatialAnchorExtension::~FSpacesSpatialAnchorExtension()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("Shutdown SpacesSpatialAnchorExtension"));
	AnchorLoadingFuture.Reset();
	AnchorSavingFuture.Reset();
	UnregisterOpenXRExtensionModularFeature();
}

const void* FSpacesSpatialAnchorExtension::OnBeginSession(XrSession InSession, const void* InNext)
{
	XRTrackingSystem = GEngine->XRSystem.Get();

	if(SnapdragonSpaces::xrCreateSpatialAnchorStoreConnectionMSFT(InSession, &AnchorStore) == false)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Failed creating the anchor store"));
		return InNext;
	}

	if(AnchorStore == XR_NULL_HANDLE)
	{
		UE_LOG(LogSpacesRuntime, Warning, TEXT("Anchor store handle is NULL"));
		return InNext;
	}
	
	return InNext;
}

void FSpacesSpatialAnchorExtension::PostCreateInstance(XrInstance InInstance)
{
	SetInstanceHandle(InInstance);
}

void FSpacesSpatialAnchorExtension::PostCreateSession(XrSession InSession)
{
	SetSessionHandle(InSession);
}

FString FSpacesSpatialAnchorExtension::GetDisplayName()
{
	return FString(TEXT("Spaces Spatial Anchor"));
}

bool FSpacesSpatialAnchorExtension::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("GetRequiredExtensions"));
	OutExtensions.Add(XR_MSFT_SPATIAL_ANCHOR_EXTENSION_NAME);
	OutExtensions.Add(XR_MSFT_SPATIAL_ANCHOR_PERSISTENCE_EXTENSION_NAME);
    return true;
}

IOpenXRCustomAnchorSupport* FSpacesSpatialAnchorExtension::GetCustomAnchorSupport()
{
	UE_LOG(LogSpacesRuntime, Log, TEXT("GetCustomAnchorSupport"));
    return this;
}

void FSpacesSpatialAnchorExtension::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
	// Checking if the async loading is finished
	if (AnchorLoadingFuture.IsValid())
	{
		UpdateAsyncLoading();
	}

	// Checking if the async saving is finished
	if (AnchorSavingFuture.IsValid())
	{
		UpdateAsyncSaving();
	}
}

bool FSpacesSpatialAnchorExtension::UpdateAsyncLoading()
{
	if (!AnchorLoadingFuture.IsReady()) return false;
	
	TMap<TWeakObjectPtr<UARPin>, TSharedPtr<FSpacesSpatialAnchor, ESPMode::ThreadSafe>> AnchorToLoad = AnchorLoadingFuture.Get();

	for(auto& Item : AnchorToLoad)
	{
		const auto Pin = Item.Key;
		if(!Pin.IsExplicitlyNull() && Pin.IsValid())
		{
			if(Item.Value.IsValid())
			{
				FSpacesSpatialAnchor* Anchor = new FSpacesSpatialAnchor();
				*Anchor = MoveTemp(*Item.Value.Get());
				if(Anchor->IsValid())
				{
					Pin->SetNativeResource(Anchor);
					Pin->OnTrackingStateChanged(EARTrackingState::Unknown);
				}
				else
				{
					Pin->OnTrackingStateChanged(EARTrackingState::StoppedTracking);
				}
			}
			else
			{
				Pin->OnTrackingStateChanged(EARTrackingState::StoppedTracking);
			}
		}
	}

	AnchorLoadingFuture.Reset();

	return true;
}

bool FSpacesSpatialAnchorExtension::UpdateAsyncSaving()
{
	if (!AnchorSavingFuture.IsReady()) return false;

	if(USpacesEngineSubsystem* SpacesSubsystem = GEngine->GetEngineSubsystem<USpacesEngineSubsystem>())
	{
		SpacesSubsystem->OnSpacesAnchorIsSaved.Broadcast(AnchorSavingFuture.Get(), AnchorSaving);
	}

	AnchorPendingToSave.Remove(AnchorSaving);
	AnchorSavingFuture.Reset();

	return !CreateAsyncSaveThread();
}

TArray<XrSpatialAnchorPersistenceNameMSFT> FSpacesSpatialAnchorExtension::GetSavedAnchorNames() const
{
	uint32_t AnchorCount = 0;
	if(SnapdragonSpaces::xrEnumeratePersistedSpatialAnchorNamesMSFT(AnchorStore, 0, &AnchorCount, nullptr) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed enumerate the names of all spatial anchors currently persisted"));
		return TArray<XrSpatialAnchorPersistenceNameMSFT>();
	}

	if(AnchorCount == 0) return TArray<XrSpatialAnchorPersistenceNameMSFT>();
	TArray<XrSpatialAnchorPersistenceNameMSFT> AnchorNames;
	AnchorNames.SetNum(AnchorCount, true);

	if(SnapdragonSpaces::xrEnumeratePersistedSpatialAnchorNamesMSFT(AnchorStore, AnchorCount, &AnchorCount, AnchorNames.GetData()) == false)
	{
		UE_LOG(LogSpacesRuntime, Error, TEXT("Failed enumerate the names of all spatial anchors currently persisted"));
		return TArray<XrSpatialAnchorPersistenceNameMSFT>();
	}

	return AnchorNames;
}

#undef LOCTEXT_NAMESPACE
